"""
This module contains a set of common lightweight wrappers to perform simple and non-customised DeltaDeltaCt analyses.
This is designed for lightweight users that do not wish to employ specialised pipelines.

> ### A Word on Plotters 
> Please, note that pipelines fully support mixing "static" and "interactive" Plotters, 
> but static figures will not stay open if interactive plotters are called to plot after them! 
> Because `qpcr.Filters` are always called to plot *before* any other `qpcr.Plotters`, this will mainly
> affect visualising the `qpcr.Plotters.ReplicateBoxplots` generated as Filter-Summaries.
"""

import qpcr
import matplotlib.pyplot as plt
import pandas as pd 
import statistics as stats
import qpcr._auxiliary.warnings as aw
import qpcr._auxiliary as aux
import qpcr.Plotters as Plotters
import qpcr.Filters as Filters
import qpcr.Readers as Readers
import re
import os 
import difflib

class Pipeline:
    """
    This is the basic template class for qpcr Pipelines. 
    It contains a set of basic preliminary methods
    that ensure that elementary required inputs are provided.

    Note
    ----
    The simplest implementation of this `Pipeline` template is the `Basic` pipeline.
    """
    def __init__(self):
        # super().__init__()
        self._Normalisers = []
        self._Assays = []
        self._save_to = None
        self._df = None
        self._stats_df = None
        self._Results = None
        self._replicates = None
        self._names = None
        self._softlink = True

    def assays(self):
        """
        Returns
        -------
        assays : list
            The linked `qpcr.Assay` objects for assays-of-interest.
        """
        assays = self._Assays
        return assays
    
    def normalisers(self):
        """
        Returns
        -------
        normalisers : list
            The linked `qpcr.Assay` objects for normaliser-assays.
        """
        normalisers = self._Normalisers
        return normalisers

    def replicates(self, replicates:(int or tuple)):
        """
        Set the replicates specifics to use for grouping.

        Parameters
        ----------
        replicates : int or tuple
            Can be an `integer` (equal group sizes, e.g. `3` for triplicates), 
            or a `tuple` (uneven group sizes, e.g. `(3,2,3)` if the second group is only a duplicate). 
        """
        self._replicates = replicates

    def names(self, names:(list or dict)):
        """
        Set names for replicates groups.

        Parameters
        ----------
        names : list or dict
            Either a `list` (new names without repetitions) or `dict` (key = old name, value = new name) specifying new group names. 
            Group names only need to be specified once, and are applied to all replicate entries.
        """
        self._names = names


    def run(self, **kwargs):
        """
        Will run the pipeline provided that at least minimal inputs have been provided (i.e. Assays and Normalisers, as well as replicate specifics have been provided).
        This is a wrapper, the actual functional core is defined in the method `self._run()`. 
        To implement your own pipeline, make sure to define your own `_run()`, redefine `run()`
        as you require.

        Parameters
        ----------
        **kwargs
            Any additional keyword arguments that will be passed to the actual `_run()` method.
        """

        # vet if there are at least one normaliser and assay present
        if self._Normalisers == [] or self._Assays == []:
            aw.HardWarning("Pipeline:no_data")

        self._run(**kwargs)
    
    def save_to(self, directory:str):
        """
        Set the location where to save result files

        Parameters
        ----------
        directory : str
            A directory to save the results to.
        """
        self._save_to = directory
        # if the directory does not yet exist, we make it
        if not os.path.exists(self._save_to):
            os.mkdir(self._save_to)

    def get(self, kind="stats"):
        """
        Returns
        -------
        data 
            A pandas dataframe either in replicate version `kind="df"`
            or in stats version `kind="stats"` (default). Alternatively,
            a qpcr.Results object can be returned using `kind="obj"`.
        """
        if kind == "stats":
            return self._stats_df
        elif kind == "df":
            df = self._df
            if "assay" in df.columns: 
                df = df.drop(columns = ["assay"])
            return df
        elif kind == "obj":
            return self._Results
    
    def results( self ):
        """
        Returns
        -------
        Results : qpcr.Results
            The `qpcr.Results` object storing the pipelines data.
            Note, this is equivalent to `get( kind = "obj" )`.
        """
        return self._Results


    def link(self, assays:(list or str) = None, normalisers:(list or str) = None):
        """
        Links new assays-of-interest / sample assays and/or normaliser assays 
        to the pipline, either replacing old ones or keeping them, 
        depending on `softlink()` settings.

        Parameters
        ----------
        assays : list or str
            A `list` of filepaths to raw datafiles of assays-of-interest, or a directory (`str`) where these are stored.
        normalisers : list or str
            A `list` of filepaths to raw datafiles of normaliser assays, or a directory (`str`) where these are stored.
        """
        if self._softlink:
            self.prune()
        self.add_assays(assays)
        self.add_normalisers(normalisers)


    def prune(self, assays = True, results = True, normalisers = True):
        """
        Will clear assays, results, and/or normalisers

        Parameters
        ----------
        assays : bool
            Will clear any sample assays in the pipeline if True (default).
        
        results : bool
            Will clear any computed results in the pipline if True (default).
        
        normalisers : bool
            Will clear any normalisers in the pipline if True (default).
        """
        if assays: self._Assays = []
        if normalisers: self._Normalisers = []
        if results: 
            self._df = None
            self._stats_df = None
            self._Results = None

    def add_normalisers(self, normalisers):
        """
        Adds normalisers (filepaths) (keeping any already present)
        
        Parameters
        ----------
        normalisers : list or str
            A `list` of filepaths to raw datafiles of normaliser assays, or a directory (`str`) where these are stored.
        """
        normalisers = self._from_directory(normalisers)
        self._Normalisers.extend(normalisers)
    
    def add_assays(self, assays):
        """
        Adds assays-of-interest / sample assays (filepaths) (keeping any already present)

        Parameters
        ----------
        assays : list or str
            A `list` of filepaths to raw datafiles of sample assays, or a directory (`str`) where these are stored.
        """
        assays = self._from_directory(assays)
        self._Assays.extend(assays)

    def softlink(self, bool = None):
        """
        If `softlink = True`, then `link_assays()` will 
        prune any previous assays. Otherwise, it will 
        add new ones and keep old ones.

        Parameters
        ----------
        bool
            Set to False to disable `softlinking` (default is True).
        """
        if bool is None:
            return self._softlink
        else:
            self._softlink = bool

    def _from_directory(self, files):
        """
        Checks if a directory was provided for assays / normalisers and returns a list of all contained files if so.
        Otherwise it just returns the list of files.
        This is used for add_assays and add_normalisers
        """
        if isinstance(files, str):

            # if inputs are a directory
            if os.path.isdir(files):
            
                # get files from the directory
                datafiles = os.listdir(files)
            
                # if no files are found, raise error
                if len(datafiles) == 0:
                    aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)
            
                # combine paths with parent directory
                datafiles = [os.path.join(files,n) for n in datafiles]
                return datafiles

            # else check if inputs are a single file
            # then just put it into a list to be valid input for .extend()
            elif os.path.isfile(files):
                
                return [files]

            else: 
                aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)
        
        # else check if we got a list or tuple of files
        elif isinstance(files, (list or tuple)):
            return files
        
        # raise error for anything else...
        else: 
            aw.HardWarning("Pipeline:no_data_input", file = files, traceback = False)


    def _run(self, **kwargs):
        """
        This is the actual function that each custom Pipeline has to define...
        """
        print("self._run() is the actual function that each custom Pipeline has to define...")


class Basic(Pipeline):
    """
    Performs simple standardized DeltaDeltaCt analysis 
    based on two lists of files, one for normaliser assays and one for sample assays.
    This makes use of standard settings for `qpcr.Analyser` and `qpcr.Normaliser`
    which cannot be customized! 
    For customization check out the `Blueprint` pipeline or generate your own.
    """
    def __init__(self):
        super().__init__()
    
    def _run(self, **kwargs):
        """
        The automated standard DeltaDeltaCt pipeline. 
        Using default settings of `qpcr.Analyser` and `qpcr.Normaliser`
        """
        reader = qpcr.DataReader()
        analyser = qpcr.Analyser()
        normaliser = qpcr.Normaliser()

        # add replicate and names info to kwargs
        kwargs = dict(kwargs, replicates = self._replicates, names = self._names)

        # analyse normalisers
        normalisers = [  reader.read(i, **kwargs) for i in self._Normalisers  ]     
        normalisers = [  analyser.pipe(i) for i in normalisers  ]
       
        # analyse sample assays
        assays = [  reader.read(i, **kwargs) for i in self._Assays  ]   
        assays = [  analyser.pipe(i) for i in assays  ]

        # now normalise
        normaliser.link( assays = assays, normalisers = normalisers )
        normaliser.normalise()

        # get and store results
        results = normaliser.get()
        self._Results = results
        self._df = results.get()
        self._stats_df = results.stats()

        if self._save_to is not None:
            results.save(self._save_to)

class BasicPlus(Basic):
    """
    The same as the Basic Pipeline, but has the option to integrate Plotters and Filters.
    """
    def __init__(self):
        super().__init__()
        self._Plotters = []
        self._Figures = []
        self._Filters = []
    
    def prune(self, assays = True, results = True, normalisers = True, figures = True, plotters = False, filters = False):
        """
        Will clear the pipeline.

        Parameters
        ----------
        assays : bool
            Will clear any sample assays in the pipeline if True (default).
        
        results : bool
            Will clear any computed results in the pipline if True (default).
        
        normalisers : bool
            Will clear any normalisers in the pipline if True (default).
        figures : bool
            Will clear any figures in the pipline if True (default).
        plotters : bool
            Will clear any Plotters from the pipeline if True (default False).
        filters : bool
            Will clear any Filters in the pipline if True (default False).
        """
        super().prune(assays = assays, results = results, normalisers = normalisers)
        if figures: 
            self._Figures = []
        if filters:
            self._Filters = []
        if plotters: 
            self._Plotters = []

    def add_plotters(self, *Plotters:object):
        """
        Adds already specified qpcr.Plotter objects to the Pipeline.
        
        Parameters
        ----------
        *Plotters
            Any number of qpcr.Plotters.Plotter objects.
        """
        self._Plotters.extend(Plotters)
    
    def add_filters(self, *Filters:object):
        """
        Adds already specified qpcr.Filter instances to the Pipeline.
        
        Parameters
        ----------
        *Filters
            Any number of `qpcr.Filters.Filter` objects.
        """
        self._Filters.extend(Filters)

    def Figures(self):
        """
        Returns
        -------
        list
            A list of all Figures generated by the pipeline's Plotters.
        """
        return self._Figures

    def _run(self, **kwargs):
        """
        The automated standard DeltaDeltaCt pipeline. 
        Also produces applies and Plotters to produce figures...
        """

        # simply set up the Blueprint pipeline 
        # and run with default settings
        pipeline = Blueprint()

        pipeline.replicates(self._replicates)
        pipeline.names(self._names)

        if self._save_to is not None:
            pipeline.save_to(self._save_to)

        pipeline.add_assays(self._Assays)
        pipeline.add_normalisers(self._Normalisers)

        pipeline.add_plotters(*self._Plotters)
        pipeline.add_filters(*self._Filters)

        pipeline.run()

        # and store the results
        self._Results = pipeline.get(kind = "obj")
        self._df = pipeline.get(kind = "df")
        self._stats_df = pipeline.get(kind = "stats")




class Blueprint(BasicPlus):
    """
    Performs simple Delta-Delta-Ct analysis based on the same workflow as the `BasicPlus` pipeline, but allows full costumization of SampleReader, Analyser, and Normaliser objects.
    Optionally, `qpcr.SampleReader`, `qpcr.Analyser`, and `qpcr.Normaliser` may be set up externally and linked into the pipeline. Any non-linked processing classes will be set up using defaults.
    """
    def __init__(self):
        super().__init__()
        self._Reader = None
        self._Analyser = None
        self._Normaliser = None

    def prune(self, assays = True, results = True, normalisers = True, figures = True, plotters = False, filters = False, cores = False, reader = False, analyser = False, normaliser = False):
        """
        Will clear the pipeline.

        Parameters
        ----------
        assays : bool
            Will clear any sample assays in the pipeline if True (default).
        
        results : bool
            Will clear any computed results in the pipline if True (default).
        
        normalisers : bool
            Will clear any normalisers in the pipline if True (default).
        figures : bool
            Will clear any figures in the pipline if True (default).
        plotters : bool
            Will clear any Plotters from the pipeline if True (default False).
        filters : bool
            Will clear any Filters in the pipline if True (default False).
        cores : bool
            Will clear Reader, Analyser, and Normaliser if True (default False).
        reader : bool
            Will only clear the Reader if True (default False).
        analyser : bool
            Will only clear the Analyser if True (default False).
        normaliser : bool
            Will only clear the Normaliser if True (default False).
            Note, that clearing `results` will also clear the `Normaliser`'s results,
            but keep the Normaliser itself!
        """
        super().prune(
                        assays = assays, 
                        results = results, 
                        normalisers = normalisers,
                        figures = figures, 
                        plotters = plotters, 
                        filters = filters
                    )
        
        try: 
            self._Normaliser.prune(
                                    assays = assays, 
                                    normalisers = normalisers, 
                                    results = results
                                )
        except: 
            pass
        
        if cores: 
            self._Reader = None
            self._Analyser = None
            self._Normaliser = None
        else:
            if reader:
                self._Reader = None
            if analyser:
                self._Analyser = None
            if normaliser:
                self._Normaliser = None

    def Reader(self, Reader : qpcr.SampleReader = None):
        """
        Links a `qpcr.SampleReader` object to the pipeline.

        Parameters
        ----------
        Reader : qpcr.SampleReader
            A `qpcr.SampleReader` object
        """
        if Reader is not None: 
            self._Reader = Reader
        return self._Reader

    def Analyser(self, Analyser : qpcr.Analyser = None):
        """
        Links a `qpcr.Analyser` object to the pipeline.

        Parameters
        ----------
        Analyser : qpcr.Analyser
            A `qpcr.Analyser` object
        """
        if Analyser is not None:
            self._Analyser = Analyser
        return self._Analyser

    def Normaliser(self, Normaliser : qpcr.Normaliser = None):
        """
        Links a `qpcr.Normaliser` object to the pipeline.

        Parameters
        ----------
        Normaliser : qpcr.Normaliser
            A `qpcr.Normaliser` object
        """
        if Normaliser is not None:
            self._Normaliser = Normaliser
        return self._Normaliser
    
    def _setup_cores(self):
        """
        Sets SampleReader, Analyser, and Normaliser to defaults, if no external ones were provided...
        """
        if self.Reader() is None: 
            self.Reader(qpcr.DataReader())
        if self.Analyser() is None: 
            self.Analyser(qpcr.Analyser())
        if self.Normaliser() is None:
            self.Normaliser(qpcr.Normaliser())

    def _run(self, **kwargs):
        """
        The automated standard DeltaDeltaCt pipeline. 
        Also produces applies and Plotters to produce figures...
        """
        # setup Analyser, and Normaliser (if none were provided)
        self._setup_cores()
        
        reader = self.Reader()
        analyser = self.Analyser()
        normaliser = self.Normaliser()

        # check if we have plotters or filters
        have_plotters = len(self._Plotters) != 0 
        have_filters = len(self._Filters) != 0

        # add replicate and names info to kwargs
        kwargs = dict(kwargs, replicates = self._replicates, names = self._names)

        # analyse normalisers
        normalisers = [  reader.read(i, **kwargs) for i in self._Normalisers  ]

        if have_filters:
            for filter in self._Filters:
                normalisers = [  filter.pipe(i) for i in normalisers  ]
            
        normalisers = [  analyser.pipe(i) for i in normalisers  ]
       
        # analyse sample assays
        assays = [  reader.read(i, **kwargs) for i in self._Assays  ]

        if have_filters:
            for filter in self._Filters:
                assays = [  filter.pipe(i) for i in assays  ]
        
        assays = [  analyser.pipe(i) for i in assays  ]

        # now normalise
        normaliser.link( assays = assays, normalisers = normalisers )
        normaliser.normalise()

        # get and store results
        results = normaliser.get()
        self._Results = results
        self._df = results.get()
        self._stats_df = results.stats()

        if self._save_to is not None:
            results.save(self._save_to)

        # plot filtering report
        if have_filters:
            for filter in self._Filters:
                if self._save_to is not None or filter.report() is not None:
                    
                    # add report location if none was specified...
                    if filter.report() is None: 
                        filter.report(self._save_to)

                # save filter summary fig
                fig = filter.plot()
                self._Figures.append(fig)

        # plot results
        if have_plotters:
            for plotter in self._Plotters:

                # generate plotter fig and save
                plotter.link(self._Results)
                fig = plotter.plot()
                self._Figures.append(fig)

                if self._save_to is not None:
                    filename = self._make_figure_filename(plotter)
                    plotter.save(filename)

    def _make_figure_filename(self, plotter):
        """
        Increments a filename with a numeric counter...
        """
        num = 1
        suffix = plotter.suffix()
        while True:
            filename = os.path.join(self._save_to, f"{plotter.id()}_{num}.{suffix}")
            if not os.path.exists(filename):
                break
            num+=1
        return filename


class _Qupid_Blueprint(Blueprint):
    """
    This is the implementation of the Plueprint pipeline for the Qupid webapp.
    It redefines the file-reading functions with methods compatible with the UploadedFile objects from streamlit.
    """
    def __init__(self):
        super().__init__()

    def _setup_cores(self):
        """
        Sets SampleReader, Analyser, and Normaliser to defaults, if no external ones were provided...
        """
        if self.Reader() is None: 
            self.Reader(qpcr._Qupid_SampleReader())
        if self.Analyser() is None: 
            self.Analyser(qpcr.Analyser())
        if self.Normaliser() is None:
            self.Normaliser(qpcr.Normaliser())


# We push this one behind in the release...
# Since we now have the ddCt pipeline we should be able to slim down most of the pipelines here anyway...

# class MultiAssay(Blueprint):
#     """
#     Performs Delta-Delta-Ct based on data from a single multi-assay datafile.
#     Datasets within this datafile must be decorated to identify them as assays-of-interest or normalisers.
#     Check out the documentation of `qpcr.Parsers` for more information on decorators.

#     Note
#     -------
#     This class relies on `qpcr.Parsers` and decorated assays to get its input data. 
#     If your file does not offer this kind of architecture, choose another pipeline.

#     """
#     def __init__(self):
#         super().__init__()
#         self._src = None
    
#     def link(self, filename : str, **kwargs):
#         """
#         Reads a datafile in csv or excel format containing 
#         multiple decorated datasets and extracts sample and normalisers assays. 

#         Parameters
#         ----------
#         filename : str
#             A filepath to a raw data file.
#         **kwargs
#             Any additional keyword arguments that shall be passed to the `qpcr.MultiReader`'s `pipe` method.
#         """
#         self._src = filename

#         # setup Reader, Analyser, and Normaliser (if none were provided)
#         self._setup_cores(**kwargs)
#         reader = self.Reader()

#         # read the multi-assay datafile
#         self._Assays, self._Normalisers = reader.pipe(self._src, **kwargs)
        
        
#     def _run(self, **kwargs):
#         """
#         The main workflow
#         """
#         # setup Reader, Analyser, and Normaliser (if none were provided)
#         self._setup_cores()
        
#         analyser = self.Analyser()
#         normaliser = self.Normaliser()

#         normalisers = []
#         samples = []

#         # analyse normalisers:
#         for norm in self._Normalisers:
#             for filter in self._Filters:
#                 norm = filter.pipe(norm)
#             norm = analyser.pipe(norm)
#             normalisers.append(norm)
#         normaliser.link(normalisers = normalisers)

#         # analyse sample assays
#         for sample in self._Assays:
#             for filter in self._Filters:
#                 sample = filter.pipe(sample)
#             sample = analyser.pipe(sample)
#             samples.append(sample)
#         normaliser.link(assays = samples)

#         normaliser.normalise()
#         results = normaliser.get()

#         self._Results = results
#         self._df = results.get()
#         self._stats_df = results.stats()

#         if self._save_to is not None:
#             results.save(self._save_to)

#         # plot filtering report
#         for filter in self._Filters:
#             if self._save_to is not None or filter.report() is not None:
#                 # add report location if none was specified...
#                 if filter.report() is None: 
#                     filter.report(self._save_to)

#             figs = filter.plot()
#             self._Figures.extend(figs)

#         # plot results
#         for plotter in self._Plotters:
#             plotter.link(self._Results)
#             fig = plotter.plot()
#             self._Figures.append(fig)

#             if self._save_to is not None:
#                 filename = self._make_figure_filename(plotter)
#                 plotter.save(filename)

#     def _setup_cores(self, **kwargs):
#         """
#         Sets Reader, Analyser, and Normaliser to defaults, if no external ones were provided...
#         """
#         # check if a sheet_name was specified in case of multi-sheet files...
#         use_multi_sheet = "sheet_name" not in kwargs
#         if self.Reader() is None: 
#             if self._is_multisheet() and use_multi_sheet:
#                 self.Reader(Readers.MultiSheetReader())
#             else: 
#                 self.Reader(Readers.MultiReader())
#         if self.Analyser() is None: 
#             self.Analyser(qpcr.Analyser())
#         if self.Normaliser() is None:
#             self.Normaliser(qpcr.Normaliser())
    
#     def _is_multisheet(self):
#         """
#         Checks if a provided excel file contains multiple sheets
#         """
#         verdict = False
#         if self._src.endswith("xlsx"):
#             data = pd.read_excel(self._src, sheet_name = None)
#             verdict = len(data.keys()) > 1
#         return verdict

#     # add_assays is disabled
#     def add_assays(self):
#         print("To provide a data input file use link()!\nIf you wish to supply separate files for assays-of-interest and normalisers, checkout another Pipeline as this one only works with a single Mlit-Assay file!")

#     # add_normalisers is disabled
#     def add_assays(self):
#         print("To provide a data input file use link()!\nIf you wish to supply separate files for assays-of-interest and normalisers, checkout another Pipeline as this one only works with a single Mlit-Assay file!")
    
class ddCt(Blueprint):
    """
    Performs only Delta-Delta-Ct and requires `qpcr.Assay` objects as inputs.
    Hence, this pipeline does NOT read any files!

    It follows the default workflow of the `BasicPlus`pipeline and is based on the `BluePrint`
    pipeline to allow customisation.

    Note
    ----
    As the pipeline inherits from the `Blueprint` pipeline it does have a `Reader` method (which won't do anything though!). 
    """
    def __init__(self):
        super().__init__()
        
    def _setup_cores(self):
        """
        Sets Analyser, and Normaliser to defaults, if no external ones were provided...
        """
        if self.Analyser() is None: 
            self.Analyser(qpcr.Analyser())
        if self.Normaliser() is None:
            self.Normaliser(qpcr.Normaliser())
        
    def _run(self, **kwargs):
        # setup Analyser, and Normaliser (if none were provided)
        self._setup_cores()
        
        analyser = self.Analyser()
        normaliser = self.Normaliser()

        # check if we have plotters or filters
        have_plotters = len(self._Plotters) != 0 
        have_filters = len(self._Filters) != 0

        # analyse normalisers
        normalisers = [  i for i in self._Normalisers  ]

        if have_filters:
            for filter in self._Filters:
                normalisers = [  filter.pipe(i) for i in normalisers  ]
            
        normalisers = [  analyser.pipe(i) for i in normalisers  ]
       
        # analyse sample assays
        assays = [  i for i in self._Assays  ]

        if have_filters:
            for filter in self._Filters:
                assays = [  filter.pipe(i) for i in assays  ]
        
        assays = [  analyser.pipe(i) for i in assays  ]

        # now normalise
        normaliser.link( assays = assays, normalisers = normalisers )
        normaliser.normalise()

        # get and store results
        results = normaliser.get()
        self._Results = results
        self._df = results.get()
        self._stats_df = results.stats()

        if self._save_to is not None:
            results.save(self._save_to)

        # plot filtering report
        if have_filters:
            for filter in self._Filters:
                if self._save_to is not None or filter.report() is not None:
                    
                    # add report location if none was specified...
                    if filter.report() is None: 
                        filter.report(self._save_to)

                # save filter summary fig
                fig = filter.plot()
                self._Figures.append(fig)

        # plot results
        if have_plotters:
            for plotter in self._Plotters:

                # generate plotter fig and save
                plotter.link(self._Results)
                fig = plotter.plot()
                self._Figures.append(fig)

                if self._save_to is not None:
                    filename = self._make_figure_filename(plotter)
                    plotter.save(filename)

if __name__ == "__main__":

    norm_files = ["./Examples/Example Data/28S.csv", "./Examples/Example Data/actin.csv"]
    sample_files = ["./Examples/Example Data/HNRNPL_nmd.csv", "./Examples/Example Data/HNRNPL_prot.csv"]

    norm_folder = "./Example Data 3/normalisers/"
    sample_folder = "./Example Data 3/samples/"

    groupnames = ["wt-", "wt+", "ko-", "ko+"]
    
    analysis = Blueprint()
    
    # grouped_analyser = qpcr.Analyser()
    # grouped_analyser.anchor("first")
    # analysis.Analyser(grouped_analyser)

    # analysis.save_to("Example Data 2")
    # analysis.add_assays(sample_files) # alternative: link() for iteratively linking new assays...
    # analysis.add_normalisers(norm_files)

    analysis.add_assays(sample_folder)
    analysis.add_normalisers(norm_folder)
    
    # print("No Reps specified, all inferred!")
    analysis.replicates(6)
    analysis.names(groupnames)

    iqr_filter = Filters.RangeFilter()
    iqr_filter.plotmode("interactive")

    iqr_filter.report("Example Data 4")
    analysis.add_filters(iqr_filter)

    preview = Plotters.PreviewResults(mode = "static")
    analysis.add_plotters(preview)

    # now that pipeline is ready, we can run!
    analysis.run()

    # now we can get results!
    results = analysis.get(kind="df")
    print(results)    
    # norm_folder = "Example Data 3/normalisers"
    # exp_folder = "Example Data 3/experimental/samples"
    # ctr_folder = "Example Data 3/experimental/samples"

    # groupnames = ["wt-", "wt+", "ko-", "ko+"]
    
    # analysis = BasicPlus(index = "Example Data 2/index.csv")
    # analysis.save_to("Example Data ")
    
    # print(analysis._index)
    # # analysis.set_experimental(assays = exp_folder, normalisers = norm_folder)
    # # analysis.set_control(assays = ctr_folder, normalisers = norm_folder)

    # # analysis.replicates(6)
    # # analysis.names(groupnames)

    # iqr_filter = Filters.RangeFilter()
    # # iqr_filter.report("Example Data 2")
    # analysis.add_filters(iqr_filter)

    # # preview = Plotters.PreviewResults(mode = "interactive")
    # # analysis.add_plotters(preview)

    # # # now that pipeline is ready, we can run!
    # # analysis.run()

    # # # now we can get results!
    # # results = analysis.get(kind="df")
    # # print(results)

    # # exit(0)



import numpy as np 
import matplotlib.pyplot as plt
import qpcr._auxiliary as aux
import seaborn as sns

def generate_palette(kwargs):
    """
    Generates a color pallete for seaborn 
    plots to support a custom color argument.
    """
    palette = aux.from_kwargs("palette", None, kwargs, rm = True)
    color = aux.from_kwargs("color", None, kwargs, rm = True)
    if color is not None: 
        try: 
            palette = sns.color_palette( color )
        except:
            palette = color
    return palette


def make_layout(df, ref_column:str):
    """
    Generates a tuple for col, rows for subplots 
    based on the distinct sets of values within ref_column.
    """
    ref_col = df[ref_column]
    ref_col = list(set(ref_col))
    ref_length = len(ref_col)

    if ref_length == 1: 
        return (1, 1)

    if ref_length % 2 == 0:
        nrows = 2
        if ref_length % 4 == 0 and ref_length > 4:
            nrows = 4
    elif ref_length % 6 == 0:
            nrows = 6
    else: 
        nrows = 3

    ncols = int(np.ceil(ref_length / nrows))
    if ncols * nrows >= ref_length:
        return ncols, nrows
    else:
        ncols = ncols+1
        return ncols, nrows

def make_layout_from_list( ref_list ):
    """
    Generates a subplot layout based on a list instead of dataframe column
    """
    ref_col = list(set(ref_list))
    ref_length = len(ref_col)

    if ref_length == 1: 
        return (1, 1)

    if ref_length % 2 == 0:
        nrows = 2
        if ref_length % 4 == 0 and ref_length > 4:
            nrows = 4
    elif ref_length % 6 == 0:
            nrows = 6
    else: 
        nrows = 3

    ncols = int(np.ceil(ref_length / nrows))
    if ncols * nrows >= ref_length:
        return ncols, nrows
    else:
        ncols = ncols+1
        return ncols, nrows

def make_speclist(maxrows, maxcols, spectype):
    """
    This function sets up the plotly speclist for subplots 
    according to the number of plots and type of figure desired
    """
    spec_list = []
    for r in range(maxrows):
        tmp = []
        for c in range(maxcols):
            tmp.append({'type':spectype})
        spec_list.append(tmp)
    return spec_list


class AxesCoords:
    """
    This class handles the coordinates for an Axes instance generated by subplots. 
    """
    def __init__(self, fig, axs, subplots:tuple):
        self._fig = fig
        self._axs = axs
        self._ncols, self._nrows = subplots
        self.col, self.row = None, None
        self._get_plot_type()
        self._init_row_and_col()
        self._coords = None

        # self._transpose = False
        
        self._first = self.col if self._type == "plt" else self.row
        self._second = self.row if self._type == "plt" else self.col
        self._limit_first = self._ncols if self._type == "plt" else self._nrows
        self._limit_second = self._nrows if self._type == "plt" else self._ncols
        
        self._check_2D()

        self._idx = 0

        self._2D_setup_ranges()

    def get(self):
        """
        Returns the current subplot coordinates
        """
        if self._is2D:
            coords = self._coords[self._idx]
        else:
            coords =  self._coords[self._idx][0]
        return coords

    def increment(self):
        """
        Switches to next subplot coordinates
        """
        self._idx += 1

    # def transpose(self, t:bool):
    #     """
    #     Switch col and row dimensions and reference attributes
    #     """    
    #     if t != self._transpose:
    #         self._transpose = t
    #         self._first = self.col if self._first == self.row else self.row
    #         self._second = self.col if self._second == self.row else self.row

    #         self._limit_first = self._ncols if self._limit_first == self._nrows else self._nrows
    #         self._limit_second = self._nrows if self._limit_second == self._ncols else self._ncols
    #         self._2D_setup_ranges()

    def subplot(self):
        """
        Returns the current subplot to be plotted (only for matplotlib figures)
        """
        coord = self.get()
        current_subplot = self._axs[coord]
        return current_subplot

    def _2D_setup_ranges(self):
        """
        Setup the possible coordinates for 2D grid
        """
        coords = []
        first_range = range(self._base, self._limit_first+self._base)
        second_range = range(self._base, self._limit_second+self._base)

        for first in first_range:
            for second in second_range:
                coords.append((first, second))
        self._coords = coords

    def _check_2D(self):
        """
        Checks if 2D coords will be required or not...
        """
        if self._type == "plotly": 
            self._is2D = True
        else:
            self._is2D = True if self._limit_second > 1 else False
        
    def _get_plot_type(self):
        """
        Sets if plotly or matplotlib figure are present, 
        and also checks if axs contain only one single subplot, 
        and makes it subscriptable if so...
        """
        fig, axs = plt.subplots()
        if type(self._fig) == type(fig):
            self._type = "plt"
        else:
            self._type = "plotly"
        
        if type(self._axs) == type(axs):
            self._axs = [self._axs]
        plt.close()

    def _setup_ranges(self):
        """
        Setup ranges of possible values
        """
        self._first_range = range(self._base, self._limit_first + self._base)
        self._second_range = range(self._base, self._limit_second + self._base)

    def _init_row_and_col(self):
        """
        Will set col and row to 1 if plotly figure is used, 
        otherwise 0 in case of matplotlib
        """
        if self._type == "plt":
            self._base = 0
        else: 
            self._base = 1
        self.col, self.row = (self._base, self._base)


import setuptools

with open("README.md", "r") as f:
    long_description = f.read()

setuptools.setup(
    name="qpcr", 
    version="3.1.4",
    author="Noah H. Kleinschmidt",
    author_email="noah.kleinschmidt@students.unibe.ch",
    description="A module to perform analysis of qPCR data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/NoahHenrikKleinschmidt/qpcr.git",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    python_requires='>=3.6',
)


"""
This module is designed for streamlined data visualisation of the qpcr generated results
It is designed to work directly with `qpcr.Results` objects. 

## `Static` vs `Interactive` Figures
----

The `Plotters` are designed to produce two kinds of figures each, either a `"static"` or  an `"interactive"` figure. 
The type of figure a specific Plotter should produce has to be specified using the `mode` argument. 

### Static Figures
_Static_  figures are made using `matplotlib` and they will open through whatever backend your matplotlib configuration 
as specified. Static figures are primarily designed for printing into labjournals and offer a greater flexibility with 
style customizibility (you can use `seaborn` styles for instance, or the matplotlib `rcparams` to style your figures). 

### Interactive Figures
_Interactive_  figures are made using `plotly` and they will open in your browser. Interactive Figures are primarily designed
for cases where your figures contain a lot of data so having a static view on them might be insufficient. Interactive figures
offer `plotly`'s native features like zooming, cropping, size-adjustments and so forth. It comes at the price of less flexibility
with regard to styling. You can set plotly `templates`, but that's about it. However, also interactive figures are perfectly 
adequate for your labjournal, and you may prefer using these for their dynamic figure size adjustments directly from your browser. 

### Plotting `kwargs` 
Both Static and Interactive Figures support a variety of keyword arguments that allow you to customise many of their 
characteristics and their underlying data handling. You can check which kwargs are passable to each type of figure in 
the documentation of each Plotter. 
"""

import qpcr.__init__ as qpcr
# import qpcr.Pipes
import qpcr._auxiliary.graphical as gx
import qpcr._auxiliary as aux 
import qpcr._auxiliary.defaults as defaults
import qpcr._auxiliary.warnings as wa
import pandas as pd 
import matplotlib.pyplot as plt
from matplotlib.lines import Line2D
import plotly.graph_objs as go
from plotly.subplots import make_subplots
import plotly
import seaborn as sns 
import numpy as np 

# setup default settings for charts

# FUTURE DROP HERE
# These are no longer refereced like this but directly from the defaults...
# we will remove them in a future version...

_default_static_PreviewBars = defaults.static_PreviewBars
_default_interactive_PreviewBars = defaults.interactive_PreviewBars

_default_static_PreviewDots = defaults.static_PreviewDots
_default_interactive_PreviewDots = defaults.interactive_PreviewDots

_default_static_ReplicateBoxPlot = defaults.static_ReplicateBoxPlot
_default_interactive_ReplicateBoxPlot = defaults.interactive_ReplicateBoxPlot

_default_static_FilterSummary = defaults.static_FilterSummary
_default_interactive_FilterSummary = defaults.interactive_FilterSummary


# get default colnames Id + Ct
raw_col_names = defaults.raw_col_names

# Concept: 
# Plotter: 
#       We define a superclass Plotter that will handle linking to data, 
#       linking to default parameters, and setting up default parameters.
#       It also provides a generic plot() method that will call on a FigureClass specific _plot() method...
#       Wether or not to use static or interactive plots is also handled by this class...
#
# Wrapper: 
#       The Wrapper is an equivalent to the Plotter superclass, but allows a facilitated user interface to 
#       several different plotters. It re-defines the public methods of the Plotter superclass and creates 
#       an instance of a specific Plotter and forwards data to that instance. 
#
# FigureClass:
#       A parent class for each type of figure. It contains two potential _plot() 
#       methods, one for interactive one for static plotting... Which one to use is decided based on the plotting mode...
#       Each FigureClass has therefore to link default parameters, then init Plotter (superclass), 
#       and define their own _static_plot() and _interactive_plot() methods! Any additionally required methods can be written as well...


class Plotter:
    """
    A superclass that handles Data Linking and Parameter setup for FigureClasses
    (not for End-User usage)

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).    
    """
    def __init__(self, mode = None):
        self._default_params = None
        self._PARAMS = {} #: _PARAMS will store the plotting parameter kwargs from both _default_params and any additional user specified parameters
        self._Results = None
        self._data = None
        self._rep_data = None
        self._id = type(self).__name__
        self._MODE = "interactive" if mode is None else mode
        self._fig = None
        self._set_plot()

    def link(self, Results:(qpcr.Results or pd.DataFrame)):
        """
        Links a Results object or pandas DataFrame of the same architecture
        as one handled by a Results object to the Plotter. 
        Note, that this will replace any previously linked data!

        Parameters
        ----------
        Results : qpcr.Results or pd.DataFrame
            A qpcr.Results Object or a pandas DataFrame of the same architecture.
        """
        
        if aux.same_type(Results, qpcr.Results()):
            self._Results = Results
            self._data = self._Results.stats()
            self._rep_data = self._Results.get()
        elif isinstance(Results, pd.DataFrame):
            self._Results = None
            self._data = Results
        else:
            wa.HardWarning("Plotter:unknown_data", obj = Results)

    def plot(self, **kwargs):
        """
        Generate Figure

        Parameters
        ----------
        **kwargs
            Any arbitrary keyword arguments to be passed to the plotting method
        
        Returns
        -------
        fig
            A figure object, either from matplotlib or plotly
        """
        total_kwargs = self.update_params(kwargs)
        fig = self._plot(**total_kwargs)
        self._fig = fig
        return fig

    def id(self, id:str = None):
        """
        Set a unique Id, by default the Classname will be used.

        Parameters
        ----------
        id : str
            A unique identifier for the Plotter object
        
        Returns
        -------
        id 
            The Plotter Id (if no id keyword was entered)
        """
        if id is not None:
            self._id = id
        else:
            return self._id

    def get(self):
        """
        Returns the DataFrame used for plotting

        Returns
        -------
        data
            A pandas DataFrame containing the data underlying the plot.
        """
        return self._data

    def params(self, **params):
        """
        Set default parameters for plotting (will be forwarded to **kwargs)
        Returns default parameters if no new parameters are added.

        Parameters
        ----------
        **params
            Any arbitrary keyword arguments to be passed on to the pre-set plotting kwargs.
            This will either set the default parameters (if none has been specified up to the function call)
            or will update the parameters. In case of keyword duplications any OLD key : value pairs will be 
            OVERWRITTEN by new ones. 
        
        Returns
        -------
        params : dict
            A dictionary of the new pre-set plotting parameters
        """
        if params != {} and self._PARAMS == {}:
            self._PARAMS = params
        elif params != {}:
            self.update_params(params, store = True)
        return self._PARAMS

    def update_params(self, kwargs, supersede = True, store = False):
        """
        Appends pre-set parameters to kwargs. 
        
        Parameters
        ----------
        kwargs : dict
            A dictionary of arbitrary keywords for plotting 
        supersede : bool 
            In case of key duplications: Old values will be replaced with new ones (if supersede = True, default), 
            or keep old ones (supersede = False).
        store : bool
            The newly set plotting parameter dictionary will be stored by the plotting object as new default (if store = True, default is False).
        
        Returns
        -------
        params : dict
            A dictionary of updated plotting parameters
        """
        if supersede:
            kwargs = dict(self._PARAMS, **kwargs)
        else:
            kwargs = dict(kwargs, **self._PARAMS)
        
        if store:
            self._PARAMS = kwargs

        return kwargs

    def save(self, filename, **kwargs):
        """
        Saves the figure to a file. 
        Figures are either saved using `plt.savefig` or `plotly.offline.plot`.

        Parameters
        ----------
        filename : str 
            A filename to save the figure to.
        
        **kwargs
            Any arbitrary keyword arguments for the respective figure saving method. 
        """
        if self._fig is None:
            wa.SoftWarning("Plotter:no_fig_yet")
        else:
            if self._MODE == "static":
                self._fig.savefig(filename, bbox_inches = 'tight', **kwargs)
            elif self._MODE == "interactive":
                plotly.offline.plot(self._fig, filename=filename, **kwargs)

    def suffix(self):
        """
        Returns
        ------- 
        suffix : str
            The appropriate file suffix for save (html for "interactive" or jpg for "static" figures)
        """
        suffix = "jpg" if self._MODE == "static" else "html"
        return suffix

    def _setup_default_params(self, static:dict, interactive:dict):
        """
        Setup and set default parameters for a FigureClass
        """
        self._static_default = static
        self._interactive_default = interactive

    def _set_plot(self):
        """
        Sets self._plot either interactive or static depending on MODe
        """
        self._plot  = self._static_plot if self._MODE == "static" else self._interactive_plot
        prev_default = self._default_params
        self._default_params = self._static_default if self._MODE == "static" else self._interactive_default

        if self.params() == {} or self.params() == prev_default:
            self.params(**self._default_params)
        else:
            self.update_params(self._default_params, store = True, supersede = False)

    def _static_plot(self, **kwargs):
        """
        The plot function that will handle static plotting (will be redefined for each FigureClass)
        """
        print("The plot function that will handle static plotting...")
        print("Surprised to see this?, \nPerhaps your desired plotting methods are not named properly. Make sure to name your method _static_plot()!")

    def _interactive_plot(self, **kwargs):
        """
        The plot function that will handle interactive plotting (will be redefined for each FigureClass)
        """
        print("The plot function that will handle interactive plotting...")
        print("Surprised to see this?, \nPerhaps your desired plotting methods are not named properly. Make sure to name your method _interactive_plot()!")

    def _prep_properties(self, kwargs):
        """
        Setup ncols, nrows, subplot titles (headers), x, y, and sterr, columns for figure...
        """
        self._setup_default_plot_cols()
        data = self.get()

        # setup reference column and figure subplots
        ref_col = aux.from_kwargs("key", "assay", kwargs, rm=True)
        ncols, nrows = aux.from_kwargs("subplots", 
                                        gx.make_layout(data, ref_col), 
                                        kwargs, rm=True
                                       )

        # transposing currently not supported...
        # if aux.from_kwargs("transpose", False, kwargs, rm = True):
        #     ncols, nrows = nrows, ncols

        headers = aux.sorted_set(data[ref_col])

        x = aux.from_kwargs("x", self._default_x, kwargs, rm = True)
        y = aux.from_kwargs("y", self._default_y, kwargs, rm = True)
        sterr = aux.from_kwargs("sterr", self._default_sterr, kwargs, rm = True)
        
        # the query to be used if group or group_name is ref_col
        query = "{ref_col} == '{q}'" if isinstance(headers[0], str) else "{ref_col} == {q}"

        return ref_col,ncols,nrows, headers, x, y, sterr, query

    def _setup_default_plot_cols(self):
        """
        Sets default columns for x and y in barcharts,
        default x
        - "group_name" column if present 
        - "group" Otherwise
        default y
        - "mean"
        default sterr
        - "stdev"
        """
        columns = self._data.columns
        self._default_x = "group_name" if "group_name" in columns else "group"
        self._default_y = "mean"
        self._default_sterr = "stdev"

    def _add_subplot_label(self, idx, subplot, start_character):
        """
        Adds A B C ... to upper left corner of a subplot...
        It will start labelling at any arbitrary start_character
        """
        subplot_label = chr(ord(start_character)+idx)
        subplot.annotate(
                    xy = (-0.1,1.03), 
                    text = subplot_label, 
                    xycoords = "axes fraction",
                    weight = "bold", fontsize = 12
                )

# The Wrapper essentially defines public methods from Plotter but 
# replaces self with self._Plotter instead...
class Wrapper:
    """
    A superclass that allows to make wrappers for multiple Plotters 
    (not for End-User usage).

    Parameters
    ----------
    kind : str
        The kind of Plotter to call. This can be any of the wrapped 
        Plotters, e.g. `kind = "GroupBars"`.   
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly). 

    """
    def __init__(self, kind : str, mode : str):

        self._Plotter = None
        self._id = type(self).__name__

        # Application of this would work as:
        # plotters = {
        #                 "some_Plotter" : some_Plotter, 
        #         }
            
        # self._Plotter = plotters[ kind ]
        # self._Plotter = self._Plotter( mode = mode )
    
    def plotter(self):
        """
        Returns
        -------
        plotter
            The currently used core Plotter
        """
        plotter = self._Plotter
        return plotter


    def link(self, Results:(qpcr.Results or pd.DataFrame)):
        """
        Links a Results object or pandas DataFrame of the same architecture
        as one handled by a Results object to the Plotter. 
        Note, that this will replace any previously linked data!

        Parameters
        ----------
        Results : qpcr.Results or pd.DataFrame
            A qpcr.Results Object or a pandas DataFrame of the same architecture.
        """
        plotter = self._Plotter
        plotter.link( Results )


    def plot(self, **kwargs):
        """
        Generate Figure

        Parameters
        ----------
        **kwargs
            Any arbitrary keyword arguments to be passed to the plotting method
        
        Returns
        -------
        fig
            A figure object, either from matplotlib or plotly
        """
        fig = self._Plotter.plot( **kwargs )
        return fig 

    def id(self, id:str = None):
        """
        Set a unique Id, by default the Classname will be used.

        Parameters
        ----------
        id : str
            A unique identifier for the Plotter object
        
        Returns
        -------
        id 
            The Plotter Id (if no id keyword was entered)
        """
        if id is not None:
            self._id = id
        else:
            return self._id

    def get(self):
        """
        Returns the DataFrame used for plotting

        Returns
        -------
        data
            A pandas DataFrame containing the data underlying the plot.
        """
        return self._Plotter.get()

    def params(self, **params):
        """
        Set default parameters for plotting (will be forwarded to **kwargs)
        Returns default parameters if no new parameters are added.

        Parameters
        ----------
        **params
            Any arbitrary keyword arguments to be passed on to the pre-set plotting kwargs.
            This will either set the default parameters (if none has been specified up to the function call)
            or will update the parameters. In case of keyword duplications any OLD key : value pairs will be 
            OVERWRITTEN by new ones. 
        
        Returns
        -------
        params : dict
            A dictionary of the new pre-set plotting parameters
        """
        params = self._Plotter.params( **params )
        return params

    def update_params(self, kwargs, supersede = True, store = False):
        """
        Appends pre-set parameters to kwargs. 
        
        Parameters
        ----------
        kwargs : dict
            A dictionary of arbitrary keywords for plotting 
        supersede : bool 
            In case of key duplications: Old values will be replaced with new ones (if supersede = True, default), 
            or keep old ones (supersede = False).
        store : bool
            The newly set plotting parameter dictionary will be stored by the plotting object as new default (if store = True, default is False).
        
        Returns
        -------
        params : dict
            A dictionary of updated plotting parameters
        """
        params = self._Plotter.update_params( kwargs, supersede = supersede, store = store )
        return params

    def save(self, filename, **kwargs):
        """
        Saves the figure to a file. 
        Figures are either saved using `plt.savefig` or `plotly.offline.plot`.

        Parameters
        ----------
        filename : str 
            A filename to save the figure to.
        
        **kwargs
            Any arbitrary keyword arguments for the respective figure saving method. 
        """
        self._Plotter.save( filename = filename, **kwargs )
    
    def suffix(self):
        """
        Returns
        ------- 
        suffix : str
            The appropriate file suffix for save (html for "interactive" or jpg for "static" figures)
        """
        suffix = self._Plotter.suffix()
        return suffix

    def id(self, id:str = None):
        """
        Set a unique Id, by default the Classname will be used.

        Parameters
        ----------
        id : str
            A unique identifier for the Plotter object
        
        Returns
        -------
        id 
            The Plotter Id (if no id keyword was entered)
        """
        if id is not None:
            self._id = id
        else:
            return self._id


class PreviewResults(Wrapper):
    """
    Generate a Preview of all results from all Assays in subplots.

    This is a wrapper for the Plotters: 

    - `AssayBars` ( which was previously `PreviewResults`, and is now the default setting )
    - `GroupBars`
    - `AssayDots`
    - `GroupDots` 

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).

    kind : str
        The kind of Plotter to call. This can be any of the four wrapped 
        Plotters, e.g. `kind = "GroupBars"`.
    """
    def __init__(self, mode : str, kind : str = "AssayBars" ):

        super().__init__( kind = kind, mode = mode )
        
        plotters = {
                        "AssayBars" : AssayBars,
                        "GroupBars" : GroupBars,
                        "AssayDots" : AssayDots,
                        "GroupDots" : GroupDots,  
                }
        
        self._Plotter = plotters[ kind ]
        self._Plotter = self._Plotter( mode = mode )


class AssayBars(Plotter):
    """
    Generate a Preview of all results from all Assays in subplots as Bar Charts.
    Each assay will be shown in a separate subplot.

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).


    Plotting Kwargs
    ----
    
    #### `"static"` Kwargs
    Static AssayBars figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's results"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Conditions"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "Mean $\Delta\Delta Ct$"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  label_subplots  : `bool`   |   Add each subplot with A, B, C ... (if True, default)   | `label_subplots = True` (default)     |
    | labeltype : `str`| The starting character for subplot labelling. By default an `"A"`. | `labeltype = "a"` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |   color : `str or list`   | The fillcolor for the individual bars   | `color = "yellow"`     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |   edgecolor : `str or list`   | The edgecolor for the individual bars   | `edgecolor = "black"`     |
    |   edgewidth : `float`   |  The width of the edge of individual bars  | `edgewidth = 0.5`     |
    |  ecolor : `str`    |   The color of errorbars   |  `ecolor = "orange"`    |
    |  **kwargs    | Any additional kwargs that can be passed to the `matplotlib`-backend pandas `.plot.bar()` API.     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive AssayBars figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's results"`    |
    | xlabel : `str`   |  The x axis label   | `xlabel = "My super qPCR samples"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = "Mean of ddCt"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    | legend_title : `str`    | The title to be displayed above the legend   |  `legend_title = "my assays"`   |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default just `"y"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Bar()`.     |      |

    """
    def __init__(self, mode:str):
        self._setup_default_params(
                                    static = defaults.static_PreviewBars, 
                                    interactive = defaults.interactive_PreviewBars
                                )
        # __init__ is going to require default_params to be already set!
        super().__init__(mode)

    def _static_plot(self, **kwargs):
        """
        The static Preview Results Figure
        """
        kwargs = self.update_params(kwargs)
        data = self.get()

        ref_col, ncols, nrows, headings, x, y, sterr, query = self._prep_properties(kwargs)       

        title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True)
        headers = aux.from_kwargs("headers", None, kwargs, rm = True)
        label_subplots = aux.from_kwargs("label_subplots", True, kwargs, rm = True)
        start_character = aux.from_kwargs("labeltype", "A", kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        rot = aux.from_kwargs("rot", None, kwargs, rm = True)

        palette = gx.generate_palette(kwargs)

        # set a seaborn style
        style = aux.from_kwargs("style", "ticks", kwargs, rm = True)
        sns.set_style( style )

        edgecolor = aux.from_kwargs("edgecolor", "white", kwargs, rm = True)
        edgewidth = aux.from_kwargs("edgewidth", None, kwargs, rm = True)

        if edgewidth is None: 
            edgewidth = aux.from_kwargs("linewidth", 1, kwargs, rm = True) 
        else: 
            aux.from_kwargs("linewidth", None, kwargs, rm = True )

        figsize = aux.from_kwargs("figsize", None, kwargs, rm = True)
        fig, axs = plt.subplots(nrows = nrows, ncols = ncols, figsize = figsize)
        fig.suptitle( title )

        # setup Coords
        Coords = gx.AxesCoords(fig, axs, (nrows, ncols))
        idx = 0
        for assay in headings: 
            try: 
                q = query.format(ref_col = ref_col, q = assay)
                tmp_df = data.query(q)
                tmp_df = tmp_df.sort_values("group")


                # now plot a new bar chart 
                subplot = Coords.subplot()

                tmp_df.plot.bar(
                                x = x, y = y, 
                                ax = subplot,
                                edgecolor = edgecolor,
                                linewidth = edgewidth,
                                color = palette,
                                **kwargs
                            )

                subplot.errorbar(
                                x = tmp_df[x], y = tmp_df[y], 
                                yerr = tmp_df[sterr], 
                                fmt = ".", markersize = 0, capsize = 3, 
                                ecolor = aux.from_kwargs("ecolor", "black", kwargs),
                            )
            
                subplot.set(
                            title = assay if headers is None else headers[idx],
                            xlabel = aux.from_kwargs("xlabel", None, kwargs),
                            ylabel = aux.from_kwargs("ylabel", "$\Delta\Delta$Ct", kwargs),        
                        )

                if rot is not None: 
                    align = "center" if rot == 0 else "left"
                    plt.setp( subplot.xaxis.get_majorticklabels(), rotation = -rot, ha=align, rotation_mode="anchor") 

                if not show_spines:
                    subplot.spines["right"].set_visible(False)
                    subplot.spines["top"].set_visible(False)
                    subplot.spines["left"].set_linewidth(1.05)
                    subplot.spines["bottom"].set_linewidth(1.05)

                # add ABCD... label to subplot
                if label_subplots:
                    self._add_subplot_label(idx, subplot, start_character)         

                Coords.increment()
                idx += 1
            except Exception as e:
                raise e 
                break
        
        plt.tight_layout()

        
        if show:
            plt.show()

        # print("<<< Static Check >>>")

        return fig 

    # def _add_subplot_label(self, idx, subplot, start_character):
    #     """
    #     Adds A B C ... to upper left corner of a subplot...
    #     It will start labelling at any arbitrary start_character
    #     """
    #     subplot_label = chr(ord(start_character)+idx)
    #     subplot.annotate(
    #                 xy = (-0.1,1.03), 
    #                 text = subplot_label, 
    #                 xycoords = "axes fraction",
    #                 weight = "bold", fontsize = 12
    #             )

    def _interactive_plot(self, **kwargs):
        """
        The interactive Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self.get()

        try: 
            # setup figure framework variables that have to be removed from 
            ref_col, ncols, nrows, headings, x, y, sterr, query = self._prep_properties(kwargs)
            headers = aux.from_kwargs("headers", headings, kwargs, rm = True)
            show = aux.from_kwargs("show", True, kwargs, rm = True)
            padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)

            if isinstance(padding, (list, tuple)):
                hpad, vpad = padding
            else: 
                hpad, vpad = padding, None

            speclist = gx.make_speclist(nrows, ncols, "xy")
            fig = make_subplots(
                                    rows = nrows, cols = ncols, 
                                    specs = speclist, 
                                    subplot_titles = headers,
                                    horizontal_spacing = hpad,
                                    vertical_spacing = vpad
                            )


            Coords = gx.AxesCoords(fig, [], (ncols, nrows))

            fig.update_layout(
                        title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True), 
                        height = aux.from_kwargs("height", None, kwargs, rm = True), 
                        width = aux.from_kwargs("width", None, kwargs, rm = True),
                        margin = {"autoexpand": True, "pad" : 0, "b": 1, "t": 50}, 
                        autosize = True, 
                        template = aux.from_kwargs("template", "plotly_white", kwargs, rm = True),
                        legend = {"title" : aux.from_kwargs("legend_title", ref_col, kwargs, rm = True),
                        },
                    )

            # set default axes labels
            xlabel = aux.from_kwargs("xlabel", "", kwargs, rm = True) 
            fig.update_xaxes(title_text = xlabel)
            
            ylabel = aux.from_kwargs("ylabel", "DeltaDeltaCt", kwargs, rm = True) 
            fig.update_yaxes(title_text = ylabel)

            idx = 0
            for assay in headings:
                row, col = Coords.get()
                tmp_df = data.query(query.format(ref_col = ref_col, q = assay))
                tmp_df = tmp_df.sort_values("group")

                # now plot a new bar chart 
                fig.add_trace(

                    go.Bar(
                        name = assay,
                        y = tmp_df[y], x = tmp_df[x], 
                        error_y=dict(type='data', array = tmp_df[sterr]),
                        hoverinfo = aux.from_kwargs("hoverinfo", "y", kwargs, rm = True), 
                        **kwargs
                    ), 

                    row, col
                )
                Coords.increment()

                idx += 1

            if show:
                fig.show()

            # print("<<< Interactive Check >>>")
            return fig 
        except Exception as e: 
            raise e 


class ReplicateBoxPlot(Plotter):
    """
    Generate a boxplot figure summary for the input sample replicates.
    

    Note
    ------
    This used to be the core of making filter summary figures. However, this has now
    been replaced by a dedicated `FilterSummary` figure class. Hence, support for linking
    `qpcr.Filters` directly to this figure class will be dropped in a future release!

    Parameters
    ----------
    Filter : qpcr.Filters.Filter
        A qpcr.Filters.Filter object. The Filter can also be set using the `filter()` method.

    mode : str
        The plotting mode (either `"interactive"` or `"static"`).

    Plotting Kwargs
    ----
    
    #### `"static"` Kwargs
    Static ReplicateBoxPlot figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    |   subplots : `tuple`   |  A tuple specifying the number of colums and rows (in that order) for the figure    | `subplots = (2, 3)`     |
    | title : `str`   |  The overall figure title   | `title = "My assays"    |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |   color : `str or list`   | The fillcolor for the boxes.   | `color = "yellow"`     |
    |  **kwargs    | Any additional kwargs that can be passed to the `seaborn`'s `boxplot()`.     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive ReplicateBoxPlot figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "My run"`    |
    | ylabel : `str`   |  The y-axis title   | `ylabel = "Raw Ct value"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Box()`.     |      |

    """
    def __init__(self, Filter = None, mode="interactive"):
        self._setup_default_params(
                                    static = defaults.static_ReplicateBoxPlot, 
                                    interactive = defaults.interactive_ReplicateBoxPlot
                                )
        # __init__ is going to require default_params to be already set!
        super().__init__(mode=mode)
        self._data = None
        if Filter is not None: 
            wa.SoftWarning("Versions:Deprecation", old = "ReplicateBoxPlot for use with Filters", new = "FilterSummary" )
        self._Filter = Filter
        self._filter_stats = None
    
    # FUTURE DROP HERE
    def filter(self, Filter):
        """
        Links a Filter object for which a report figure shall be generated

        Note
        -------
        Support for linking Filters to use this Figure class to visualse Filter Summaries will be 
        dropped in a future release! Use the dedicated `FilterSummary` figure class instead!

        Parameters
        ----------
        Filter : qpcr.Filters.Filter
            A qpcr.Filters.Filter object.

        """
        wa.SoftWarning("Versions:Deprecation", old = "ReplicateBoxPlot for use with Filters", new = "FilterSummary" )
        self._Filter = Filter

    def clear(self):
        """
        Will clear the currently stored Assay data
        """
        self._data = None

    def link(self, Assay:qpcr.Assay):
        """
        Links an Assay object to the BoxPlotter.
        This will simply add the Ct column to the current overall data!
        Hence, repeated linking of Assay objects will add data and NOT replace any existing.

        Parameters
        ----------
        Assay : qpcr.Assay
            A qpcr.Assay object.
        """
        if aux.same_type(Assay, qpcr.Assay()):
            self._Results = Assay
            data = self._Results.get( copy = True )
        else:
            wa.HardWarning("Plotter:unknown_data", obj = Assay)

        # add itentifier column
        data["assay"] = [self._Results.id() for i in range(len(data))]

        if self._data is None: 
            self._data = data
        else:
            self._data = pd.concat([self._data, data], ignore_index=True)
        

    def _interactive_plot(self, **kwargs):
        """
        Generates an interactive Boxplot summary of the input Ct values
        """
        kwargs = self.update_params(kwargs)

        show = aux.from_kwargs("show", True, kwargs, rm = True)
        template = aux.from_kwargs("template", None, kwargs, rm=True)
        title = aux.from_kwargs("title", None, kwargs, rm=True)

        data = self._data
    
        groups = aux.sorted_set(data["group"])
        group_names = aux.sorted_set(data["group_name"])

        fig = go.Figure()

        fig.update_layout(
                            boxmode="group", 
                            height = aux.from_kwargs("height", None, kwargs, rm = True), 
                            width = aux.from_kwargs("width", None, kwargs, rm = True),
                            template = template, 
                            title = title,
                        )

        # add default ylabel
        ylabel = aux.from_kwargs("ylabel", "Ct", kwargs, rm = True) 
        fig.update_yaxes(title_text = ylabel)

        fig.update_xaxes(showgrid=True)

        for group, name in zip(groups, group_names):
            tmp_df = data.query(f"group == {group}")
            
            
            fig.add_trace(
                            go.Box(
                                x = tmp_df["assay"],
                                y = tmp_df["Ct"],
                                name = name,
                                hoverinfo = "y+name",
                                **kwargs,
                            ),
                        )

        if show:
            fig.show()

        return fig 


    def _static_plot(self, **kwargs):
        """
        Generates a static Boxplot summary of the input Ct values
        """
        kwargs = self.update_params(kwargs)
        data = self._data
        
        # set a seaborn style
        style = aux.from_kwargs("style", "ticks", kwargs, rm = True)
        sns.set_style( style )

        show = aux.from_kwargs("show", True, kwargs, rm = True)
        figsize = aux.from_kwargs("figsize", (7,5), kwargs, rm=True)
        title = aux.from_kwargs("title", None, kwargs, rm=True)
        palette = gx.generate_palette(kwargs)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        ncols, nrows = aux.from_kwargs("subplots", gx.make_layout(data, "assay"), kwargs, rm=True)

        fig, axs = plt.subplots(nrows = nrows, ncols = ncols, figsize=figsize)
        
        # possibly nrows and ncols should be switched...
        Coords = gx.AxesCoords(fig, axs, (nrows, ncols))
        
        for assay in aux.sorted_set(data["assay"]):
            
            try: 
                ax = Coords.subplot()
            except: 
                print("We ran out of axes ... ")
                break
            tmp = data.query(f"assay == '{assay}'")

            sns.boxplot(
                        data=tmp, 
                        x = "assay", y = "Ct", hue="group_name", 
                        palette = palette, 
                        ax = ax, 
                        **kwargs
                    )

            ax.legend(bbox_to_anchor=(1,1), loc = None).remove()
            ax.set(title=assay, xlabel = "", xticklabels = [],)
            
            if not show_spines:
                sns.despine()

            Coords.increment()

        # add one single legend to the last plot
        ax.legend(bbox_to_anchor=(1,1), loc = None)

        fig.suptitle(title)
        plt.tight_layout()

        if show:
            fig.show()

        return fig 


class FilterSummary(Plotter):
    """
    Generates a summary figure of the replicate
    Ct values for each assay before and after filtering.

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).


    
    #### `"static"` Kwargs
    Static FilterSummary figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    |   subplots : `tuple`   |  A tuple specifying the number of colums and rows (in that order) for the figure    | `subplots = (2, 3)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's run"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Conditions"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "My Ct values"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |   color : `str or list`   | The color for the boxes.   | `color = ["yellow", "green"]`     |
    |  **kwargs    | Any additional kwargs that can be passed to the `seaborn`'s `boxplot()`.     |      |
  
    <br></br>
    #### `"interactive"` Kwargs
    Interactive FilterSummary figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's run"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = "Raw Ct values"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  colors : `list`   |   List of two colors for _before_ and _after_ filtering boxes. | `colors = ["red", "green"]`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |  headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default `"y+x+name"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Box()`.     |      |


    """
    def __init__(self, mode):
        self._setup_default_params(
                                    static = defaults.static_FilterSummary,
                                    interactive = defaults.interactive_FilterSummary
                                )
        super().__init__( mode = mode)
        self._before = qpcr.Results()
        self._after = qpcr.Results()

    def clear(self):
        """
        Clears the pre- and post-filtering Ct value records.
        """
        self._before = qpcr.Results()
        self._after = qpcr.Results()

    def add_before(self, assay : qpcr.Assay ):
        """
        Add a pre-filtered set of Ct values
        from an `qpcr.Assay` object. 

        Parameters
        ----------
        assay : qpcr.Assay
            An `qpcr.Assay` object
        """
        # setup groups and stuff information 
        if self._before.is_empty():
            self._before.adopt_names( assay )
            self._after.adopt_names( assay )

        self._before.add_Ct( assay )
    
    def add_after(self, assay : qpcr.Assay ):
        """
        Add a post-filtered set of Ct values
        from an `qpcr.Assay` object. 

        Parameters
        ----------
        assay : qpcr.Assay
            An `qpcr.Assay` object
        """
        self._after.add_Ct( assay )


    def _interactive_plot(self, **kwargs):
        """
        Generates an interactive Filter Summary Figure
        """
        kwargs = self.update_params(kwargs)

        show = aux.from_kwargs("show", True, kwargs, rm = True)
        template = aux.from_kwargs("template", None, kwargs, rm=True)
        title = aux.from_kwargs("title", None, kwargs, rm=True)
        padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)
        
        # setup color scheme
        colors = aux.from_kwargs("colors", ["blue", "crimson"], kwargs, rm = True)
        before_color, after_color = colors

        if isinstance(padding, (list, tuple)):
            hpad, vpad = padding
        else: 
            hpad, vpad = padding, None

        # get the data
        before = self._before.get()
        after = self._after.get()

        # get the number of assays to visualse
        assays = [ i for i in before.columns if i not in [ "group", "group_name", raw_col_names[0]] ]
        headers = aux.from_kwargs("headers", assays, kwargs, rm = True)

        # each assay will be on its own plot,
        ncols, nrows = gx.make_layout_from_list( assays )
        
        speclist = gx.make_speclist(nrows, ncols, "xy")
        fig = make_subplots(
                                rows = nrows, cols = ncols, 
                                specs = speclist, 
                                subplot_titles = headers,
                                horizontal_spacing = hpad,
                                vertical_spacing = vpad,
                                # sharex = True
                        )
        
        fig.update_layout(
                            boxmode="group", 
                            height = aux.from_kwargs("height", None, kwargs, rm = True), 
                            width = aux.from_kwargs("width", None, kwargs, rm = True),
                            template = template, 
                            title = title,
                        )

        # add default ylabel
        ylabel = aux.from_kwargs("ylabel", "Ct", kwargs, rm = True) 
        fig.update_yaxes(title_text = ylabel)
        
        Coords = gx.AxesCoords(fig, [], (ncols, nrows))

        show_legend = True
        # now iterate over each assay and make a BoxPlot
        for assay in assays : 
            
            # get before and after filter datasets for each assay
            pre_Cts = before[ ["group", "group_name", assay] ]
            post_Cts = after[ ["group", "group_name", assay] ]
            
            row, col = Coords.get()

            # pre-filter boxes 
            fig.add_trace(
                            go.Box(
                                x = pre_Cts["group_name"],
                                y = pre_Cts[ assay ],
                                name = "Before",
                                hoverinfo = "y+x+name",
                                marker_color = before_color,
                                legendgroup='group1',
                                showlegend = show_legend,
                                **kwargs,
                            ),
                            row, col
                        )
            # post-filter boxes
            fig.add_trace(
                            go.Box(
                                x = post_Cts["group_name"],
                                y = post_Cts[ assay ],
                                name = "After",
                                hoverinfo = "y+x+name",
                                marker_color = after_color,
                                legendgroup='group2',
                                showlegend = show_legend,
                                **kwargs,
                            ),
                            row, col
                        )

            Coords.increment()

            # set show_legend to False after the first assay,
            # to avoid repetitive legends...
            show_legend = False

        fig.update_layout( showlegend = True )
        if show:
            fig.show()

        return fig 


    def _static_plot(self, **kwargs):
        """
        Generates a static Filter Summary Fig
        """
        kwargs = self.update_params(kwargs)

        show = aux.from_kwargs("show", True, kwargs, rm = True)
        figsize = aux.from_kwargs("figsize", (7,5), kwargs, rm=True)
        title = aux.from_kwargs("title", None, kwargs, rm=True)
        ylabel = aux.from_kwargs("ylabel", "Ct", kwargs, rm = True)
        xlabel = aux.from_kwargs("xlabel", "", kwargs, rm = True)
        rot = aux.from_kwargs("rot",None, kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        palette = gx.generate_palette(kwargs)

        # set a seaborn style
        style = aux.from_kwargs("style", "ticks", kwargs, rm = True)
        sns.set_style( style )

        # get the data
        before = self._before.get()
        after = self._after.get()

        # get the number of assays to visualse
        assays = [ i for i in before.columns if i not in [ "group", "group_name", raw_col_names[0]] ]
        headers = aux.from_kwargs("headers", assays, kwargs, rm = True)

        # each assay will be on its own plot
        ncols, nrows = gx.make_layout_from_list( assays )

        ncols, nrows = aux.from_kwargs("subplots", (ncols, nrows), kwargs, rm=True)


        fig, axs = plt.subplots(nrows = nrows, ncols = ncols, figsize=figsize)
        
        # possibly nrows and ncols should be switched...
        Coords = gx.AxesCoords(fig, axs, (nrows, ncols))

        # set a kind column for grouping in the boxplots
        before[ "kind" ] = "before" 
        after[ "kind" ] = "after"

        show_legend = True
        for assay, header in zip( assays, headers ): 

            # get before and after filter datasets for each assay
            pre_Cts = before[ ["group", "group_name", "kind", assay] ]
            post_Cts = after[ ["group", "group_name", "kind", assay] ]

            # assemble data to one dataframe
            df = pd.concat( (pre_Cts, post_Cts) )

            # get the new subplot
            try: 
                ax = Coords.subplot()
            except: 
                print("We ran out of axes ... ")
                break

            # main box plot
            sns.boxplot(
                        data = df, 
                        x = "group_name",
                        y = assay,
                        hue = "kind",
                        palette = palette, 
                        ax = ax,
                        **kwargs
                    )
            
            # format the subplot
            ax.set(title = header, ylabel = ylabel, xlabel = xlabel)
            if rot is not None and rot != 0 :
                plt.setp( ax.xaxis.get_majorticklabels(), rotation = -rot, ha="left", rotation_mode="anchor") 
            if not show_legend: 
                ax.legend().remove()
            else:
                ax.legend( bbox_to_anchor=(1,1), loc = None, frameon = False )
            show_legend = False

            Coords.increment()

        if not show_spines: 
            sns.despine()

        fig.suptitle(title)

        plt.tight_layout()
        if show: 
            plt.show()

        return fig


class AssayDots(Plotter):
    """
    Generate a Preview of all results from all Assays in subplots, plotting individual values
    in a Dot Plot rather than Bar Plot. Each assay will be shown in a separate subplot.

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).


    Plotting Kwargs
    ----
    
    #### `"static"` Kwargs
    Static AssayDots figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's results"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Conditions"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "Mean $\Delta\Delta Ct$"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  label_subplots  : `bool`   |   Add each subplot with A, B, C ... (if True, default)   | `label_subplots = True` (default)     |
    | labeltype : `str`| The starting character for subplot labelling. By default an `"A"`. | `labeltype = "a"` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |   violin   : `bool` |  Show symmetric kde of the dots of each group (if True).   | `violin = False`    |
    |   color : `str or list`   | The fillcolor for the individual dots from replicate groups   | `color = "yellow"`     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |  **kwargs    | Any additional kwargs that can be passed to the `seaborn`'s `stripplot`.     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive AssayDots figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's results"`    |
    | xlabel : `str`   |  The x axis label   | `xlabel = "My super qPCR samples"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = "Mean of ddCt"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |   violin   : `bool` |  Show symmetric kde of the dots of each group (if True).   | `violin = False`    |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default just `"y"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Violin()`.     |      |

    """
    def __init__(self, mode:str):
        self._setup_default_params(
                                    static = defaults.static_PreviewDots, 
                                    interactive = defaults.interactive_PreviewDots
                                )
        # __init__ is going to require default_params to be already set!
        super().__init__(mode)

    def _static_plot(self, **kwargs):
        """
        The static Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self._rep_data

        # get assays to plot
        assays = [ i for i in data.columns if i not in [ "group", "group_name", raw_col_names[0] ] ]

        # generate subplot layout
        ncols, nrows = gx.make_layout_from_list( assays )

        # get kwargs incompatible with the main plotting method
        title = aux.from_kwargs("title", None, kwargs, rm = True)
        headers = aux.from_kwargs("headers", assays, kwargs, rm = True)
        label_subplots = aux.from_kwargs("label_subplots", True, kwargs, rm = True)
        start_character = aux.from_kwargs("labeltype", "A", kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        show_violins = aux.from_kwargs("violin", True, kwargs, rm = True)
        rot = aux.from_kwargs("rot", None, kwargs, rm = True)
        alpha = aux.from_kwargs("alpha", 1, kwargs, rm = True)
        xlabel = aux.from_kwargs("xlabel", None, kwargs, rm = True)
        ylabel = aux.from_kwargs("ylabel", None, kwargs, rm = True)

        # generate a custom color palette in case color kwarg is provided
        palette = gx.generate_palette(kwargs)

        # set a seaborn style
        style = aux.from_kwargs("style", "dark", kwargs, rm = True)
        sns.set_style( style )

        # make figure
        figsize = aux.from_kwargs("figsize", None, kwargs, rm = True)
        fig, axs = plt.subplots(nrows = nrows, ncols = ncols, figsize = figsize)
        fig.suptitle( title )

        # setup Coords
        Coords = gx.AxesCoords(fig, axs, (nrows, ncols))
        idx = 0
        for assay in assays: 

            try: 
                tmp_df = data[ ["group", "group_name", assay] ]
                tmp_df = tmp_df.sort_values("group")

                # now plot a new violin chart 
                subplot = Coords.subplot()

                if show_violins:
                    sns.violinplot(
                                    x = tmp_df["group_name"],
                                    y = tmp_df[ assay ],
                                    color = None,
                                    inner = None, 
                                    palette = palette,
                                    ax = subplot,

                                )
                    for i in subplot.collections:
                        i.set_alpha( alpha * 0.3 )

                sns.stripplot(
                                x = tmp_df["group_name"],
                                y = tmp_df[ assay ],
                                palette = palette,
                                alpha = alpha,
                                ax = subplot,
                                **kwargs
                            )
            
                subplot.set(
                            title = assay if headers is None else headers[idx],
                            xlabel = xlabel,
                            ylabel = ylabel,        
                        )

                # adjust xtick rotation   
                if rot is not None: 
                    align = "center" if rot == 0 else "left"
                    plt.setp( subplot.xaxis.get_majorticklabels(), rotation = -rot, ha=align, rotation_mode="anchor") 

                if not show_spines:
                    subplot.spines["right"].set_visible(False)
                    subplot.spines["top"].set_visible(False)
                    subplot.spines["left"].set_linewidth(1.05)
                    subplot.spines["bottom"].set_linewidth(1.05)

                # add ABCD... label to subplot
                if label_subplots:
                    self._add_subplot_label(idx, subplot, start_character)         

                Coords.increment()
                idx += 1
            except Exception as e:
                raise e 
                break
        
        plt.tight_layout()

        if show:
            plt.show()

        # print("<<< Static Check >>>")

        return fig 

    def _add_subplot_label(self, idx, subplot, start_character):
        """
        Adds A B C ... to upper left corner of a subplot...
        It will start labelling at any arbitrary start_character
        """
        subplot_label = chr(ord(start_character)+idx)
        subplot.annotate(
                            xy = (-0.1,1.03), 
                            text = subplot_label, 
                            xycoords = "axes fraction",
                            weight = "bold", fontsize = 12
                        )

    def _interactive_plot(self, **kwargs):
        """
        The interactive Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self._rep_data

        # get assays to plot
        assays = [ i for i in data.columns if i not in [ "group", "group_name", raw_col_names[0] ] ]
        
        # get the info of groups present in the data
        # for later use as xtick labels
        group_names = data[ ["group", "group_name"] ]
        group_names = group_names.sort_values( "group" )
        group_names = aux.sorted_set( group_names[ "group_name"] )
        ticks = np.arange( len(group_names) )

        # make subplot layout
        ncols, nrows = gx.make_layout_from_list( assays )

        try: 
            # get incompaltible kwargs 
            headers = aux.from_kwargs("headers", assays, kwargs, rm = True)
            show = aux.from_kwargs("show", True, kwargs, rm = True)
            show_violins = aux.from_kwargs("violin", True, kwargs, rm = True)

            # setup padding
            padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)
            if isinstance(padding, (list, tuple)):
                hpad, vpad = padding
            else: 
                hpad, vpad = padding, None

            # make figure
            speclist = gx.make_speclist(nrows, ncols, "xy")
            fig = make_subplots(
                                    rows = nrows, cols = ncols, 
                                    specs = speclist, 
                                    subplot_titles = headers,
                                    horizontal_spacing = hpad,
                                    vertical_spacing = vpad
                            )

            # setup subplot coords handling
            Coords = gx.AxesCoords(fig, [], (ncols, nrows))

            # setup figure
            fig.update_layout(
                                title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True), 
                                height = aux.from_kwargs("height", None, kwargs, rm = True), 
                                width = aux.from_kwargs("width", None, kwargs, rm = True),
                                margin = {"autoexpand": True, "pad" : 0, "b": 1, "t": 50}, 
                                autosize = True, 
                                template = aux.from_kwargs("template", "plotly_white", kwargs, rm = True),
                                # legend = {  "title" : aux.from_kwargs("legend_title", "Assay", kwargs, rm = True),  },
                                showlegend = False
                            )

            # set default axes labels
            xlabel = aux.from_kwargs("xlabel", "", kwargs, rm = True) 
            fig.update_xaxes(title_text = xlabel)
            
            ylabel = aux.from_kwargs("ylabel", "", kwargs, rm = True) 
            fig.update_yaxes(title_text = ylabel)

            idx = 0
            for assay, header in zip( assays, headers ):
                
                row, col = Coords.get()

                tmp_df = data[ ["group", "group_name", assay] ]
                tmp_df = tmp_df.sort_values("group")

                # now plot a new violin chart 
                fig.add_trace(

                    go.Violin(
                                    name = header,
                                    y = tmp_df[ assay ], x = tmp_df[ "group" ], 
                                    points = "all",
                                    pointpos = 0,
                                    hoverinfo = aux.from_kwargs("hoverinfo", "y", kwargs, rm = True), 
                                    **kwargs
                            ), 
                                    row, col
                        )

                # remove violins if not desired (leaving only dot plot)
                if not show_violins: 
                    fig.update_traces(
                                        fillcolor="rgba(0,0,0,0)", 
                                        line_width = 0, 
                                        selector=dict(type='violin'),
                                    )


                # update x axis to categorical group names
                fig.update_layout(
                                    { 
                                        f"xaxis{idx+1}" : dict(
                                                                tickmode = 'array',
                                                                tickvals = ticks,
                                                                ticktext = group_names,
                                                        )         
                                    }
                                )

                Coords.increment()

                idx += 1

            if show:
                fig.show()

            # print("<<< Interactive Check >>>")
            return fig 
        except Exception as e: 
            raise e 


class GroupBars(Plotter):
    """
    Generates a Bar plot figure with a separate subplot for each group. 

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).

    Plotting Kwargs
    ----
    
    #### `"static"` Kwargs
    Static GroupBars figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's results"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Conditions"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "Mean $\Delta\Delta Ct$"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  label_subplots  : `bool`   |   Add each subplot with A, B, C ... (if True, default)   | `label_subplots = True` (default)     |
    | labeltype : `str`| The starting character for subplot labelling. By default an `"A"`. | `labeltype = "a"` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |   color : `str or list`   | The fillcolor for the individual bars   | `color = "yellow"`     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |   edgecolor : `str or list`   | The edgecolor for the individual bars   | `edgecolor = "black"`     |
    |   edgewidth : `float`   |  The width of the edge of individual bars  | `edgewidth = 0.5`     |
    |  ecolor : `str`    |   The color of errorbars   |  `ecolor = "orange"`    |
    |  **kwargs    | Any additional kwargs that can be passed to the `matplotlib`-backend pandas `.plot.bar()` API.     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive GroupBars figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's results"`    |
    | xlabel : `str`   |  The x axis label   | `xlabel = "My super qPCR samples"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = "Mean of ddCt"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    | legend_title : `str`    | The title to be displayed above the legend   |  `legend_title = "my assays"`   |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default just `"y"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Bar()`.     |      |

    """
    def __init__(self, mode : str):
        self._setup_default_params(
                                    static = defaults.static_PreviewBars, 
                                    interactive = defaults.interactive_PreviewBars
                                )
        super().__init__( mode = mode )
    
    
    def _static_plot(self, **kwargs):
        """
        Generate a static figure
        """

        kwargs = self.update_params(kwargs)
        data = self.get()
        groups = data["group"].unique()
        names = data["group_name"].unique()

        ref_col, ncols, nrows, headings, x, y, sterr, query = self._prep_properties(kwargs)
        x = "assay" 

        title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True)
        headers = aux.from_kwargs("headers", names, kwargs, rm = True)
        label_subplots = aux.from_kwargs("label_subplots", True, kwargs, rm = True)
        start_character = aux.from_kwargs("labeltype", "A", kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        rot = aux.from_kwargs("rot", None, kwargs, rm = True)

        palette = gx.generate_palette(kwargs)

        xlabel = aux.from_kwargs("xlabel", None, kwargs, rm = True)
        ylabel = aux.from_kwargs("ylabel", "$\Delta\Delta$Ct", kwargs, rm = True)
        # set a seaborn style
        style = aux.from_kwargs("style", "ticks", kwargs, rm = True)
        sns.set_style( style )

        edgecolor = aux.from_kwargs("edgecolor", "white", kwargs, rm = True)
        edgewidth = aux.from_kwargs("linewidth", 1, kwargs, rm = True)
        edgewidth = aux.from_kwargs("edgewidth", edgewidth, kwargs, rm = True)
        figsize = aux.from_kwargs("figsize", None, kwargs, rm = True)

        groups = data["group"].unique()
        nrows, ncols = gx.make_layout_from_list( groups )
        
        fig, axs = plt.subplots( ncols, nrows, figsize = figsize )
        fig.suptitle( title )

        coords = gx.AxesCoords( fig, axs, (ncols, nrows) )

        idx = 0 
        for group, name in zip(groups, headers): 

            ax = coords.subplot()
            
            tmp_df = data.query( f"group == {group}" )
            tmp_df = tmp_df.sort_values( x )

            tmp_df.plot.bar(
                        x = x, 
                        y = y,
                        # yerr = "stdev",
                        color = palette,
                        edgecolor = edgecolor,
                        linewidth = edgewidth,
                        ax = ax,
                        **kwargs
                    )

            ax.errorbar(
                            x = tmp_df[x], y = tmp_df[y], 
                            yerr = tmp_df[sterr], 
                            fmt = ".", markersize = 0, capsize = 3, 
                            ecolor = aux.from_kwargs("ecolor", "black", kwargs),
                    )
            
            ax.set(
                            title = name,
                            xlabel = xlabel,
                            ylabel = ylabel,        
                )

            if rot is not None: 
                align = "center" if rot == 0 else "left"
                plt.setp( ax.xaxis.get_majorticklabels(), rotation = -rot, ha=align, rotation_mode="anchor") 


            if not show_spines:
                ax.spines["right"].set_visible(False)
                ax.spines["top"].set_visible(False)
                ax.spines["left"].set_linewidth(1.05)
                ax.spines["bottom"].set_linewidth(1.05)

                # add ABCD... label to subplot
                if label_subplots:
                    self._add_subplot_label(idx, ax, start_character)         

            coords.increment()
            idx += 1

        plt.tight_layout()

        if show:
            plt.show()
        
        return fig 


    def _interactive_plot(self, **kwargs):
        """
        The interactive Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self.get()
        groups = data["group"].unique()
        names = data["group_name"].unique()

        try: 
            # setup figure framework variables that have to be removed from 
            ref_col, ncols, nrows, headings, x, y, sterr, query = self._prep_properties(kwargs)
            x = "assay"
            headers = aux.from_kwargs("headers", names, kwargs, rm = True)
            show = aux.from_kwargs("show", True, kwargs, rm = True)
            padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)

            if isinstance(padding, (list, tuple)):
                hpad, vpad = padding
            else: 
                hpad, vpad = padding, None

            # make new ncols and nrows based on the groups list
            nrows, ncols = gx.make_layout_from_list(groups)

            speclist = gx.make_speclist(nrows, ncols, "xy")
            fig = make_subplots(
                                    rows = nrows, cols = ncols, 
                                    specs = speclist, 
                                    subplot_titles = headers,
                                    horizontal_spacing = hpad,
                                    vertical_spacing = vpad
                            )


            Coords = gx.AxesCoords(fig, [], (ncols, nrows))

            fig.update_layout(
                        title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True), 
                        height = aux.from_kwargs("height", None, kwargs, rm = True), 
                        width = aux.from_kwargs("width", None, kwargs, rm = True),
                        margin = {"autoexpand": True, "pad" : 0, "b": 1, "t": 50}, 
                        autosize = True, 
                        template = aux.from_kwargs("template", "plotly_white", kwargs, rm = True),
                        legend = {"title" : aux.from_kwargs("legend_title", "group name", kwargs, rm = True),
                        },
                    )

            # set default axes labels
            xlabel = aux.from_kwargs("xlabel", "", kwargs, rm = True) 
            fig.update_xaxes(title_text = xlabel)
            
            ylabel = aux.from_kwargs("ylabel", "DeltaDeltaCt", kwargs, rm = True) 
            fig.update_yaxes(title_text = ylabel)

            idx = 0
            for group, name in zip( groups, names ):
                row, col = Coords.get()
                tmp_df = data.query(f"group == {group}")
                tmp_df = tmp_df.sort_values( x )

                # now plot a new bar chart 
                fig.add_trace(

                    go.Bar(
                        name = name,
                        y = tmp_df[y], x = tmp_df[x], 
                        error_y=dict(type='data', array = tmp_df[sterr]),
                        hoverinfo = aux.from_kwargs("hoverinfo", "y", kwargs, rm = True), 
                        **kwargs
                    ), 

                    row, col
                )
                Coords.increment()

                idx += 1

            if show:
                fig.show()

            # print("<<< Interactive Check >>>")
            return fig 
        except Exception as e: 
            raise e 





class GroupDots(Plotter):
    """
    Generate a Preview of all results from all Assays in subplots, plotting individual values
    in a Dot Plot rather than Bar Plot.

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).


    Plotting Kwargs
    ----
    
    #### `"static"` Kwargs
    Static GroupDots figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's results"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Conditions"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "Mean $\Delta\Delta Ct$"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  label_subplots  : `bool`   |   Add each subplot with A, B, C ... (if True, default)   | `label_subplots = True` (default)     |
    | labeltype : `str`| The starting character for subplot labelling. By default an `"A"`. | `labeltype = "a"` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |   violin   : `bool` |  Show symmetric kde of the dots of each group (if True).   | `violin = False`    |
    |   color : `str or list`   | The fillcolor for the individual dots from replicate groups   | `color = "yellow"`     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |  **kwargs    | Any additional kwargs that can be passed to the `seaborn`'s `stripplot()`.     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive GroupDots figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's results"`    |
    | xlabel : `str`   |  The x axis label   | `xlabel = "My super qPCR samples"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = "Mean of ddCt"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |   violin   : `bool` |  Show symmetric kde of the dots of each group (if True).   | `violin = False`    |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default just `"y"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Violin()`.     |      |

    """
    def __init__(self, mode:str):
        self._setup_default_params(
                                    static = defaults.static_PreviewDots, 
                                    interactive = defaults.interactive_PreviewDots
                                )
        # __init__ is going to require default_params to be already set!
        super().__init__(mode)

    def _static_plot(self, **kwargs):
        """
        The static Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self._rep_data

        # get assays to plot
        assays = [ i for i in data.columns if i not in [ "group", "group_name", raw_col_names[0] ] ]
        assays = sorted( assays )

        # get the groups
        groups = data["group"].unique()
        names = data["group_name"].unique()
        
        # generate subplot layout
        ncols, nrows = gx.make_layout_from_list( groups )

        # get kwargs incompatible with the main plotting method
        title = aux.from_kwargs("title", None, kwargs, rm = True)
        headers = aux.from_kwargs("headers", names, kwargs, rm = True)
        label_subplots = aux.from_kwargs("label_subplots", True, kwargs, rm = True)
        start_character = aux.from_kwargs("labeltype", "A", kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        show_violins = aux.from_kwargs("violin", True, kwargs, rm = True)
        rot = aux.from_kwargs("rot", None, kwargs, rm = True)
        alpha = aux.from_kwargs("alpha", 1, kwargs, rm = True)
        xlabel = aux.from_kwargs("xlabel", None, kwargs, rm = True)
        ylabel = aux.from_kwargs("ylabel", None, kwargs, rm = True)

        # generate a custom color palette in case color kwarg is provided
        palette = gx.generate_palette(kwargs)

        # set a seaborn style
        style = aux.from_kwargs("style", "dark", kwargs, rm = True)
        sns.set_style( style )

        # make figure
        figsize = aux.from_kwargs("figsize", None, kwargs, rm = True)
        fig, axs = plt.subplots(nrows = nrows, ncols = ncols, figsize = figsize)
        fig.suptitle( title )

        # setup Coords
        Coords = gx.AxesCoords(fig, axs, (nrows, ncols))
        idx = 0
        for group, name in zip( groups, headers ): 

            try:
                
                # prepare a vertical bigtable format dataframe 
                tmp_df = self._prepare_df(data, assays, group)
                
                # now plot a new violin chart 
                subplot = Coords.subplot()

                if show_violins:
                    sns.violinplot(
                                    x = tmp_df[ "assay" ],
                                    y = tmp_df[ "value" ],
                                    color = None,
                                    inner = None, 
                                    palette = palette,
                                    ax = subplot,

                                )
                    for i in subplot.collections:
                        i.set_alpha( alpha * 0.3 )

                sns.stripplot(
                                x = tmp_df[ "assay" ],
                                y = tmp_df[ "value" ],
                                palette = palette,
                                alpha = alpha,
                                ax = subplot,
                                **kwargs
                            )
            
                subplot.set(
                            title = name,
                            xlabel = xlabel,
                            ylabel = ylabel,        
                        )

                # adjust xtick rotation   
                if rot is not None: 
                    align = "center" if rot == 0 else "left"
                    plt.setp( subplot.xaxis.get_majorticklabels(), rotation = -rot, ha=align, rotation_mode="anchor") 

                if not show_spines:
                    subplot.spines["right"].set_visible(False)
                    subplot.spines["top"].set_visible(False)
                    subplot.spines["left"].set_linewidth(1.05)
                    subplot.spines["bottom"].set_linewidth(1.05)

                # add ABCD... label to subplot
                if label_subplots:
                    self._add_subplot_label(idx, subplot, start_character)         

                Coords.increment()
                idx += 1
            except Exception as e:
                raise e 
                break
        
        plt.tight_layout()

        if show:
            plt.show()

        # print("<<< Static Check >>>")

        return fig 

    def _prepare_df(self, data, assays, group):
        """
        Concatenates the different assay colums into a single 
        set of two columns, one for all "value"s and one for 
        the "assay" identifiers. 
        """
        # get a group subset 
        group_df = data.query( f"group == {group}" )

        # concatenate all separate assay columns into a set of 
        # 'value' and 'assay' columns.

        # setup the _assays concatenated dataframe   
        _assays = group_df[ [ "group", assays[0] ] ]
        _assays = _assays.rename( columns = { assays[0] : "value" } )
        _assays["assay"] = [ assays[0] for i in _assays["group"] ]

        # now iteratively add all remaining assays
        for assay in assays[1:]:
            tmp_df = group_df[ [ "group", assay ] ]
            tmp_df = tmp_df.rename( columns = { assay : "value" } )
            tmp_df["assay"] = [ assay for i in tmp_df["group"] ]
            _assays = pd.concat( [_assays, tmp_df], ignore_index = True)
                
        # remove the group col, and sort
        tmp_df = _assays.drop( columns = ["group"] )
        tmp_df = tmp_df.sort_values( "assay" )

        return tmp_df

    def _add_subplot_label(self, idx, subplot, start_character):
        """
        Adds A B C ... to upper left corner of a subplot...
        It will start labelling at any arbitrary start_character
        """
        subplot_label = chr(ord(start_character)+idx)
        subplot.annotate(
                            xy = (-0.1,1.03), 
                            text = subplot_label, 
                            xycoords = "axes fraction",
                            weight = "bold", fontsize = 12
                        )

    def _interactive_plot(self, **kwargs):
        """
        The interactive Preview Results Figure
        """

        kwargs = self.update_params(kwargs)
        data = self._rep_data

        # get assays to plot
        assays = [ i for i in data.columns if i not in [ "group", "group_name", raw_col_names[0] ] ]
        assays = sorted( assays )

        # get the groups
        groups = data["group"].unique()
        names = data["group_name"].unique()
        ticks = np.arange( len(assays) )

        # make subplot layout
        ncols, nrows = gx.make_layout_from_list( groups )

        try: 
            # get incompaltible kwargs 
            headers = aux.from_kwargs("headers", names, kwargs, rm = True)
            show = aux.from_kwargs("show", True, kwargs, rm = True)
            show_violins = aux.from_kwargs("violin", True, kwargs, rm = True)

            # setup padding
            padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)
            if isinstance(padding, (list, tuple)):
                hpad, vpad = padding
            else: 
                hpad, vpad = padding, None

            # make figure
            speclist = gx.make_speclist(nrows, ncols, "xy")
            fig = make_subplots(
                                    rows = nrows, cols = ncols, 
                                    specs = speclist, 
                                    subplot_titles = headers,
                                    horizontal_spacing = hpad,
                                    vertical_spacing = vpad
                            )

            # setup subplot coords handling
            Coords = gx.AxesCoords(fig, [], (ncols, nrows))

            # setup figure
            fig.update_layout(
                                title = aux.from_kwargs("title", "Results Preview", kwargs, rm = True), 
                                height = aux.from_kwargs("height", None, kwargs, rm = True), 
                                width = aux.from_kwargs("width", None, kwargs, rm = True),
                                margin = {"autoexpand": True, "pad" : 0, "b": 1, "t": 50}, 
                                autosize = True, 
                                template = aux.from_kwargs("template", "plotly_white", kwargs, rm = True),
                                # legend = {  "title" : aux.from_kwargs("legend_title", "Assay", kwargs, rm = True),  },
                                showlegend = False
                            )

            # set default axes labels
            xlabel = aux.from_kwargs("xlabel", "", kwargs, rm = True) 
            fig.update_xaxes(title_text = xlabel)
            
            ylabel = aux.from_kwargs("ylabel", "", kwargs, rm = True) 
            fig.update_yaxes(title_text = ylabel)

            idx = 0
            for group, name in zip( groups, headers ):
                
                # prepare a vertical bigtable format dataframe 
                tmp_df = self._prepare_df(data, assays, group)

                row, col = Coords.get()

                # now plot a new violin chart 
                fig.add_trace(

                    go.Violin(
                                    name = name,
                                    x = tmp_df[ "assay" ], 
                                    y = tmp_df[ "value" ], 
                                    points = "all",
                                    pointpos = 0,
                                    hoverinfo = aux.from_kwargs("hoverinfo", "y", kwargs, rm = True), 
                                    **kwargs
                            ), 
                                    row, col
                        )

                # remove violins if not desired (leaving only dot plot)
                if not show_violins: 
                    fig.update_traces(
                                        fillcolor="rgba(0,0,0,0)", 
                                        line_width = 0, 
                                        selector=dict(type='violin'),
                                    )


                # update x axis to categorical group names
                fig.update_layout(
                                    { 
                                        f"xaxis{idx+1}" : dict(
                                                                tickmode = 'array',
                                                                tickvals = ticks,
                                                                ticktext = assays,
                                                        )         
                                    }
                                )

                Coords.increment()

                idx += 1

            if show:
                fig.show()

            # print("<<< Interactive Check >>>")
            return fig 
        except Exception as e: 
            raise e 


class EfficiencyLines(Plotter):
    """
    Generates a Figure for the linear regressions used for Assay efficiency 
    calculations. This FigureClass specifically works with the `qpcr.Calibrator` 
    class. 

    Parameters
    ----------
    mode : str
        The plotting mode. May be either "static" (matplotlib) or "interactive" (plotly).
    
    Plotting Kwargs
    ----

    #### `"static"` Kwargs
    Static EfficiencyLines figures accept the following kwargs:
    
    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    |   figsize : `tuple`   |  The figure size    | `figsize = (10, 4)`     |
    | title : `str`   |  The overall figure title   | `title = "Today's efficiencies"    |
    | xlabel : `str`| The x-axis label of each subplot | `xlabel = "Log Dilution"` |
    | ylabel : `str`| The y-axis label of each subplot | `ylabel = "My Ct Values"` |
    | rot : `float` | The rotation of x-axis labels | `rot = 0.3` |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  label_subplots  : `bool`   |   Add each subplot with A, B, C ... (if True, default)   | `label_subplots = True` (default)     |
    | labeltype : `str`| The starting character for subplot labelling. By default an `"A"`. | `labeltype = "a"` |
    |   frame   : `bool` |  Show left and top spines of subplots (if True)    | `frame = False` (default)     |
    |   color : `str or list`   | The fillcolor for the individual bars   | `color = "yellow"`     |
    |  style : `str`   | A `seaborn` style to set. Check out available styles [here](https://www.python-graph-gallery.com/104-seaborn-themes).     | `style = "darkgrid"`    |
    |   edgecolor : `str or list`   | The edgecolor for the individual dots for the datapoints.   | `edgecolor = "black"`     |
    |   edgewidth : `float`   |  The width of the edge of individual dots.  | `edgewidth = 0.5`     |
    |   linecolor : `str or list`   | The color for regression line.   | `linecolor = "crimson"`     |
    |   linewidth : `float`   |  The width of the regression line.  | `edgewidth = 0.5`     |
    |  **kwargs    | Any additional kwargs that can be passed to `seaborn`'s `scatterplot` and `lineplot` (both!).     |      |

    <br></br>
    #### `"interactive"` Kwargs
    Interactive EfficiencyLines figures accept the following kwargs:

    |   Argument  |  Description    |  Example    |
    | ---- | ---- | ---- |
    |  show : `bool`    |  Whether or not to show the figure    |  `show = True` (default)   |
    | title : `str`   |  The overall figure title   | `title = "Today's efficiencies"`    |
    | xlabel : `str`   |  The x axis label   | `xlabel = "Log Dilution"`    |
    | ylabel : `str`   |  The y axis label   | `ylabel = My super Ct values"`    |
    |  height : `int`   |   Height of the figure   | `height = 50`    |
    |  width : `int`   |   Width of the figure   | `width = 50`    |
    |  padding : `float or tuple`   |   Padding between subplots. This can be a single float (interpreted as horizontal padding), or a tuple of (horizontal, vertical) paddings.   | `padding = 0.2`    |
    |  template : `str`   | The `plotly` template to use. Check out available templates [here](https://plotly.com/python/templates/).     | `template = "plotly_dark"`    |
    |    headers : `list` |  A list of titles for each subplot in the preview figure    | `headers = ["transcript A", "transcript B"]`     |
    |  hoverinfo : `str`   | The type of hoverinfo to display. By default just `"y+x"`. Learn more about plotly hoverinfo [here](https://plotly.com/python/hover-text-and-formatting/). Please, note that `hovertemplate` is not currently supported.  | `hoverinfo = "name+y"`    |
    |  **kwargs    | Any additional kwargs that can be passed to `plotly`'s`graphs_objs.Scatter()`.     |      |


    """
    def __init__(self, mode : str ):
        self._setup_default_params(
                                    static = defaults.static_EfficiencyLines, 
                                    interactive = defaults.interactive_EfficiencyLines
                                )
        super().__init__( mode = mode )
        self._Calibrator = None

    def link( self, calibrator : qpcr.Calibrator ):
        """
        Links a `qpcr.Calibrator` object to source data from.

        Note
        -----
        Only a `qpcr.Calibrator` that has actually *de novo*
        computed efficiencies will have data to plot, and only for 
        the newly computed effiencies! 

        Parameters
        ----------
        calibrator : qpcr.Calibrator
            A `qpcr.Calibrator` object that has computed new efficiencies.
        """
        self._Calibrator = calibrator
    
    def _static_plot(self, **kwargs):
        """
        Generates a static Effiency Regression Figure
        """
        kwargs = self.update_params(kwargs)
        data = self._Calibrator._computed_values
        headers = list(  data.keys()  )

        title = aux.from_kwargs("title", None, kwargs, rm = True)
        headers = aux.from_kwargs("headers", headers, kwargs, rm = True)
        label_subplots = aux.from_kwargs("label_subplots", True, kwargs, rm = True)
        start_character = aux.from_kwargs("labeltype", "A", kwargs, rm = True)
        show_spines = aux.from_kwargs("frame", True, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        rot = aux.from_kwargs("rot", None, kwargs, rm = True)

        palette = gx.generate_palette(kwargs)

        xlabel = aux.from_kwargs("xlabel", None, kwargs, rm = True)
        ylabel = aux.from_kwargs("ylabel", None, kwargs, rm = True)

        # set a seaborn style
        style = aux.from_kwargs("style", "ticks", kwargs, rm = True)
        sns.set_style( style )

        edgecolor = aux.from_kwargs("edgecolor", "white", kwargs, rm = True)
        edgewidth = aux.from_kwargs("edgewidth", None, kwargs, rm = True)

        linecolor = aux.from_kwargs("linecolor", "black", kwargs, rm = True)
        linewidth = aux.from_kwargs("linewidth", 1, kwargs, rm = True)

        ncols, nrows = gx.make_layout_from_list( headers )

        fig, axs = plt.subplots( nrows = nrows, ncols = ncols)
        fig.suptitle( title )
        Coords = gx.AxesCoords( fig, axs, subplots = (nrows, ncols) )

        idx = 0 
        for id, obj in data.items(): 
            
            ax = Coords.subplot()

            # get data to plot              
            dilutions, cts = obj.values()
            eff = obj.efficiency()
            model = obj.model()
            Rsquare = model.rvalue ** 2

            # plot the regression line
            yvals = model.slope * dilutions + model.intercept
            sns.lineplot(
                                x = dilutions,
                                y = yvals,
                                color = linecolor,
                                linewidth = linewidth,
                                ax = ax,
                                **kwargs
                    )

            # plot the original data
            sns.scatterplot(
                                x = dilutions,
                                y = cts,
                                palette = palette, 
                                edgecolor = edgecolor,
                                linewidth = edgewidth,
                                ax = ax,
                                **kwargs
                            )

            # add additional info to a legend
            ax.legend( handles = [
                                    # R^2 Value
                                    Line2D(  [0], [0], 
                                            color = "black", 
                                            visible = False, 
                                            label = f"$R^2$ \t = {Rsquare:.4f}"
                                        ),
                                    # Efficiency Value
                                    Line2D(  [0], [0], 
                                            color = "black", 
                                            visible = False, 
                                            label = f"$eff.$\t= {eff:.4f}"
                                        )
                                ], 
                                loc = "upper right",
                                frameon = False 
                    )

            # some formatting...
            ax.set(
                    title = id, 
                    xlabel = xlabel,
                    ylabel = ylabel,
                )

            if rot is not None: 
                align = "center" if rot == 0 else "left"
                plt.setp( ax.xaxis.get_majorticklabels(), rotation = -rot, ha=align, rotation_mode="anchor") 

            # add ABCD... label to subplot
            if label_subplots:
                self._add_subplot_label(idx, ax, start_character)         

            if not show_spines:
                sns.despine()
            
            idx += 1
            Coords.increment()

        plt.tight_layout()
        if show: 
            plt.show()

        return fig 

    def _interactive_plot(self, **kwargs):
        """
        Generates an interactive EfficiencyLines figure
        """
        kwargs = self.update_params(kwargs)
        data = self._Calibrator._computed_values
        headers = list(  data.keys()  )

        title = aux.from_kwargs("title", None, kwargs, rm = True)
        headers = aux.from_kwargs("headers", headers, kwargs, rm = True)
        show = aux.from_kwargs("show", True, kwargs, rm = True)
        padding = aux.from_kwargs("padding", 0.1, kwargs, rm = True)

        if isinstance(padding, (list, tuple)):
            hpad, vpad = padding
        else: 
            hpad, vpad = padding, None

        # make new ncols and nrows based on the headers list
        nrows, ncols = gx.make_layout_from_list(headers)

        speclist = gx.make_speclist(nrows, ncols, "xy")
        fig = make_subplots(
                                rows = nrows, cols = ncols, 
                                specs = speclist, 
                                subplot_titles = headers,
                                horizontal_spacing = hpad,
                                vertical_spacing = vpad
                        )


        Coords = gx.AxesCoords(fig, [], (ncols, nrows))

        fig.update_layout(
                    title = title, 
                    height = aux.from_kwargs("height", None, kwargs, rm = True), 
                    width = aux.from_kwargs("width", None, kwargs, rm = True),
                    margin = {"autoexpand": True, "pad" : 0, "b": 1, "t": 50}, 
                    autosize = True, 
                    template = aux.from_kwargs("template", "plotly_white", kwargs, rm = True),
                    # legend = {"title" : aux.from_kwargs("legend_title", "Assay", kwargs, rm = True), },
                )

        # set default axes labels
        xlabel = aux.from_kwargs("xlabel", None, kwargs, rm = True) 
        fig.update_xaxes(title_text = xlabel)
        
        ylabel = aux.from_kwargs("ylabel", None, kwargs, rm = True) 
        fig.update_yaxes(title_text = ylabel)

        Coords = gx.AxesCoords(fig, [], (ncols, nrows))

        # iterate over the data...
        idx = 0 
        for id, obj in data.items():

            row, col = Coords.get()

            # get data to plot              
            dilutions, cts = obj.values()
            eff = obj.efficiency()
            model = obj.model()
            Rsquare = model.rvalue ** 2

            # plot the original data
            fig.add_trace(
                            go.Scatter(
                                name = id,
                                x = dilutions,
                                y = cts,  
                                mode = "markers",
                                showlegend = False,
                                hoverinfo = aux.from_kwargs("hoverinfo", "y+x", kwargs, rm = True), 
                                **kwargs
                            ), 
                            row, col
                        )
            
            # plot the regression line
            yvals = model.slope * dilutions + model.intercept

            fig.add_trace(
                            go.Scatter(
                                name = id,
                                x = dilutions,
                                y = yvals,  
                                mode = "lines",
                                showlegend = False,
                                **kwargs
                            ), 
                            row, col
                        )


            # add infos as annotations (because custom legends don't work here)
            # NOTE: The xref x{idx} is a pretty nice lifehack to essentially 
            #       emulate the ax behaviour from matplotib.subplots... 
            xref = "x" if idx == 0 else f"x{idx+1}"
            yref = "y" if idx == 0 else f"y{idx+1}"
            info_text = f"R^2 = {Rsquare:.4f}<br>eff. = {eff:.4f}"

            fig.add_annotation(
                            dict(
                                    x = 0, 
                                    y = np.max( cts ), 
                                    xref = xref, 
                                    yref = yref, 
                                    text = info_text, 
                                    showarrow = False
                                )
                            )

            Coords.increment()
            idx += 1

        if show: 
            fig.show()

        return fig 



# if __name__ == '__main__':

    # files = ["Example Data/28S.csv", "Example Data/28S_again.csv", "Example Data/actin.csv",
    # "Example Data/actin2.csv",
    # "Example Data/actin3.csv",
    # "Example Data/HNRNPL_nmd.csv", "Example Data/HNRNPL_prot.csv"]
    # groupnames = ["wt-", "wt+", "ko-", "ko+"]

    # # setup figures
    # a = PreviewResults(mode = "static")
    # b = PreviewResults(mode = "interactive")

    # a.params(
    #       frame = True, labeltype = "a", show = False
    # )
    # b.params(
    #     show = False, template = "plotly"  
    # )

    # # predefined pipeline use
    # pipe = qpcr.Pipes.Basic()
    # pipe.link(files)
    # pipe.add_normalisers(files[:2])
    # pipe.replicates(6)
    # pipe.names(groupnames)


    # print()
    # print("==== Multiple (ALL) Samples ====")
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot(color = "green")
    # except: 
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot()
    # except: print("Interactive Failed!")

    # print()
    # print("==== Multiple (7) Samples ====")
    # pipe.link(files[:7])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot(show = False)
    # except: 
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot(template = "plotly")
    # except: print("Interactive Failed!")

    # print()
    # print("==== Multiple (6) Samples ====")
    # pipe.link(files[:6])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot()
    # except: 
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot()
    # except: print("Interactive Failed!")

    # print()
    # print("==== Multiple (4) Samples ====")
    # pipe.link(files[:4])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot()
    # except: 
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot()
    # except: print("Interactive Failed!")

    # print()
    # print("==== Multiple (3) Samples ====")
    # pipe.link(files[:3])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot()
    # except Exception as e: 
    #     print(e)
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot()
    # except Exception as e: 
    #     print("Interactive Failed!")
    #     raise e

    # print()
    # print("==== Multiple (2) Samples ====")
    # pipe.link(files[:2])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot(show = False)
    # except Exception as e: 
    #     print(e)
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot(show = False)
    # except Exception as e: 
    #     print("Interactive Failed!")
    #     raise e


    # print()
    # print("==== One (1) Sample ====")
    # pipe.link([files[0]])
    # pipe.run()
    # result = pipe.get(kind = "obj")
    # try:
    #     a.link(result)
    #     a.plot()
    # except: 
    #     print("Static Failed!")

    # try:
    #     b.link(result)
    #     b.plot()
    # except: print("Interactive Failed!")
    # exit(0)